"""
Definition of the `fiftyone` command-line interface (CLI).

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import argparse
from collections import defaultdict
from datetime import datetime
import json
import os
import subprocess
import sys
import time
import textwrap

import argcomplete
from tabulate import tabulate
import webbrowser

import eta.core.serial as etas
import eta.core.utils as etau

import fiftyone as fo
import fiftyone.constants as foc
import fiftyone.core.config as focg
import fiftyone.core.dataset as fod
import fiftyone.core.session as fos
import fiftyone.core.utils as fou
import fiftyone.migrations as fom
import fiftyone.utils.data as foud
import fiftyone.utils.image as foui
import fiftyone.utils.quickstart as fouq
import fiftyone.utils.video as fouv
import fiftyone.zoo.datasets as fozd
import fiftyone.zoo.models as fozm


_TABLE_FORMAT = "simple"
_MAX_CONSTANT_VALUE_COL_WIDTH = 79


class Command(object):
    """Interface for defining commands.

    Command instances must implement the `setup()` method, and they should
    implement the `execute()` method if they perform any functionality beyond
    defining subparsers.
    """

    @staticmethod
    def setup(parser):
        """Setup the command-line arguments for the command.

        Args:
            parser: an `argparse.ArgumentParser` instance
        """
        raise NotImplementedError("subclass must implement setup()")

    @staticmethod
    def execute(parser, args):
        """Executes the command on the given args.

        args:
            parser: the `argparse.ArgumentParser` instance for the command
            args: an `argparse.Namespace` instance containing the arguments
                for the command
        """
        raise NotImplementedError("subclass must implement execute()")


class FiftyOneCommand(Command):
    """The FiftyOne command-line interface."""

    @staticmethod
    def setup(parser):
        subparsers = parser.add_subparsers(title="available commands")
        _register_command(subparsers, "quickstart", QuickstartCommand)
        _register_command(subparsers, "annotation", AnnotationCommand)
        _register_command(subparsers, "app", AppCommand)
        _register_command(subparsers, "config", ConfigCommand)
        _register_command(subparsers, "constants", ConstantsCommand)
        _register_command(subparsers, "convert", ConvertCommand)
        _register_command(subparsers, "datasets", DatasetsCommand)
        _register_command(subparsers, "migrate", MigrateCommand)
        _register_command(subparsers, "utils", UtilsCommand)
        _register_command(subparsers, "zoo", ZooCommand)

    @staticmethod
    def execute(parser, args):
        parser.print_help()


class QuickstartCommand(Command):
    """Launch a FiftyOne quickstart.

    Examples::

        # Launch the quickstart
        fiftyone quickstart

        # Launch the quickstart with a video dataset
        fiftyone quickstart --video

        # Launch the quickstart as a remote session
        fiftyone quickstart --remote

        # Launch the quickstart in a desktop App session
        fiftyone quickstart --desktop
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "-v",
            "--video",
            action="store_true",
            help="launch the quickstart with a video dataset",
        )
        parser.add_argument(
            "-p",
            "--port",
            metavar="PORT",
            default=None,
            type=int,
            help="the port number to use",
        )
        parser.add_argument(
            "-A",
            "--address",
            metavar="ADDRESS",
            default=None,
            type=str,
            help="the address (server name) to use",
        )
        parser.add_argument(
            "-r",
            "--remote",
            action="store_true",
            help="whether to launch a remote App session",
        )
        parser.add_argument(
            "-a",
            "--desktop",
            action="store_true",
            help="whether to launch a desktop App instance",
        )
        parser.add_argument(
            "-w",
            "--wait",
            metavar="WAIT",
            default=3,
            type=float,
            help=(
                "the number of seconds to wait for a new App connection "
                "before returning if all connections are lost. If negative, "
                "the process will wait forever, regardless of connections"
            ),
        )

    @staticmethod
    def execute(parser, args):
        # If desktop wasn't explicitly requested, fallback to default
        desktop = args.desktop or None

        _, session = fouq.quickstart(
            video=args.video,
            port=args.port,
            address=args.address,
            remote=args.remote,
            desktop=desktop,
        )

        _watch_session(session, args.wait)


class ConfigCommand(Command):
    """Tools for working with your FiftyOne config.

    Examples::

        # Print your entire config
        fiftyone config

        # Print a specific config field
        fiftyone config <field>

        # Print the location of your config on disk (if one exists)
        fiftyone config --locate
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "field", nargs="?", metavar="FIELD", help="a config field to print"
        )
        parser.add_argument(
            "-l",
            "--locate",
            action="store_true",
            help="print the location of your config on disk",
        )

    @staticmethod
    def execute(parser, args):
        if args.locate:
            config_path = focg.locate_config()
            if os.path.isfile(config_path):
                print(config_path)
            else:
                print("No config file found at '%s'" % config_path)

            return

        if args.field:
            field = getattr(fo.config, args.field)
            if etau.is_str(field):
                print(field)
            else:
                print(etas.json_to_str(field))
        else:
            print(fo.config)


class ConstantsCommand(Command):
    """Print constants from `fiftyone.constants`.

    Examples::

        # Print all constants
        fiftyone constants

        # Print a specific constant
        fiftyone constants <CONSTANT>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "constant",
            nargs="?",
            metavar="CONSTANT",
            help="the constant to print",
        )

    @staticmethod
    def execute(parser, args):
        if args.constant:
            print(getattr(foc, args.constant))
            return

        # Print all constants
        _print_constants_table(
            {
                k: v
                for k, v in vars(foc).items()
                if not k.startswith("_") and k == k.upper()
            }
        )


def _print_constants_table(d):
    contents = sorted(
        ((k, _render_constant_value(v)) for k, v in d.items()),
        key=lambda kv: kv[0],
    )
    table_str = tabulate(
        contents, headers=["constant", "value"], tablefmt=_TABLE_FORMAT
    )
    print(table_str)


def _render_constant_value(value):
    value = str(value)
    if (
        _MAX_CONSTANT_VALUE_COL_WIDTH is not None
        and len(value) > _MAX_CONSTANT_VALUE_COL_WIDTH
    ):
        value = value[: (_MAX_CONSTANT_VALUE_COL_WIDTH - 4)] + " ..."

    return value


class ConvertCommand(Command):
    """Convert datasets on disk between supported formats.

    Examples::

        # Convert an image classification directory tree to TFRecords format
        fiftyone convert \\
            --input-dir /path/to/image-classification-directory-tree \\
            --input-type fiftyone.types.ImageClassificationDirectoryTree \\
            --output-dir /path/for/tf-image-classification-dataset \\
            --output-type fiftyone.types.TFImageClassificationDataset

        # Convert a COCO detection dataset to CVAT image format
        fiftyone convert \\
            --input-dir /path/to/coco-detection-dataset \\
            --input-type fiftyone.types.COCODetectionDataset \\
            --output-dir /path/for/cvat-image-dataset \\
            --output-type fiftyone.types.CVATImageDataset

        # Perform a customized conversion via optional kwargs
        fiftyone convert \\
            --input-dir /path/to/coco-detection-dataset \\
            --input-type fiftyone.types.COCODetectionDataset \\
            --input-kwargs max_samples=100 shuffle=True \\
            --output-dir /path/for/cvat-image-dataset \\
            --output-type fiftyone.types.TFObjectDetectionDataset \\
            --output-kwargs force_rgb=True \\
            --overwrite
    """

    @staticmethod
    def setup(parser):
        required = parser.add_argument_group("required arguments")
        required.add_argument(
            "--input-type",
            metavar="INPUT_TYPE",
            help="the fiftyone.types.Dataset type of the input dataset",
            required=True,
        )
        required.add_argument(
            "--output-type",
            metavar="OUTPUT_TYPE",
            help="the fiftyone.types.Dataset type to output",
            required=True,
        )

        parser.add_argument(
            "--input-dir",
            metavar="INPUT_DIR",
            help="the directory containing the dataset",
        )
        parser.add_argument(
            "--input-kwargs",
            nargs="+",
            metavar="KEY=VAL",
            action=_ParseKwargsAction,
            help=(
                "additional keyword arguments for "
                "`fiftyone.utils.data.convert_dataset(..., input_kwargs=)`"
            ),
        )
        parser.add_argument(
            "--output-dir",
            metavar="OUTPUT_DIR",
            help="the directory to which to write the output dataset",
        )
        parser.add_argument(
            "--output-kwargs",
            nargs="+",
            metavar="KEY=VAL",
            action=_ParseKwargsAction,
            help=(
                "additional keyword arguments for "
                "`fiftyone.utils.data.convert_dataset(..., output_kwargs=)`"
            ),
        )
        parser.add_argument(
            "-o",
            "--overwrite",
            action="store_true",
            help="whether to overwrite an existing output directory",
        )

    @staticmethod
    def execute(parser, args):
        foud.convert_dataset(
            input_dir=args.input_dir,
            input_type=etau.get_class(args.input_type),
            input_kwargs=args.input_kwargs,
            output_dir=args.output_dir,
            output_type=etau.get_class(args.output_type),
            output_kwargs=args.output_kwargs,
            overwrite=args.overwrite,
        )


class DatasetsCommand(Command):
    """Tools for working with FiftyOne datasets."""

    @staticmethod
    def setup(parser):
        subparsers = parser.add_subparsers(title="available commands")
        _register_command(subparsers, "list", DatasetsListCommand)
        _register_command(subparsers, "info", DatasetsInfoCommand)
        _register_command(subparsers, "stats", DatasetsStatsCommand)
        _register_command(subparsers, "create", DatasetsCreateCommand)
        _register_command(subparsers, "head", DatasetsHeadCommand)
        _register_command(subparsers, "tail", DatasetsTailCommand)
        _register_command(subparsers, "stream", DatasetsStreamCommand)
        _register_command(subparsers, "export", DatasetsExportCommand)
        _register_command(subparsers, "draw", DatasetsDrawCommand)
        _register_command(subparsers, "rename", DatasetsRenameCommand)
        _register_command(subparsers, "delete", DatasetsDeleteCommand)

    @staticmethod
    def execute(parser, args):
        parser.print_help()


class DatasetsListCommand(Command):
    """List FiftyOne datasets.

    Examples::

        # List available datasets
        fiftyone datasets list
    """

    @staticmethod
    def setup(parser):
        pass

    @staticmethod
    def execute(parser, args):
        datasets = fod.list_datasets()

        if datasets:
            for dataset in datasets:
                print(dataset)
        else:
            print("No datasets found")


class DatasetsInfoCommand(Command):
    """Print information about FiftyOne datasets.

    Examples::

        # Print basic information about all datasets
        fiftyone datasets info
        fiftyone datasets info --sort-by created_at
        fiftyone datasets info --sort-by name --reverse

        # Print information about a specific dataset
        fiftyone datasets info <name>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name",
            nargs="?",
            metavar="NAME",
            help="the name of a dataset",
        )
        parser.add_argument(
            "-s",
            "--sort-by",
            metavar="FIELD",
            default="last_loaded_at",
            help="a field to sort the dataset rows by",
        )
        parser.add_argument(
            "-r",
            "--reverse",
            action="store_true",
            help="whether to print the results in reverse order",
        )

    @staticmethod
    def execute(parser, args):
        if args.name:
            _print_dataset_info(args.name)
        else:
            _print_all_dataset_info(args.sort_by, args.reverse)


def _print_dataset_info(name):
    dataset = fod.load_dataset(name)
    print(dataset)


def _print_all_dataset_info(sort_by, reverse):
    info = fod.list_datasets(info=True)

    headers = [
        "name",
        "created_at",
        "last_loaded_at",
        "version",
        "persistent",
        "media_type",
        "tags",
        "num_samples",
    ]

    if sort_by in headers:
        key = lambda d: (d[sort_by] is not None, d[sort_by])
        info = sorted(info, key=key, reverse=not reverse)
    else:
        print("Ignoring invalid sort-by field '%s'" % sort_by)

    records = [tuple(_format_cell(i[key]) for key in headers) for i in info]

    table_str = tabulate(records, headers=headers, tablefmt=_TABLE_FORMAT)
    print(table_str)


def _format_cell(cell):
    if cell == True:
        return "\u2713"

    if cell == False:
        return ""

    if cell is None:
        return "???"

    if isinstance(cell, datetime):
        return cell.replace(microsecond=0)

    if etau.is_container(cell):
        return ",".join(cell)

    return cell


class DatasetsStatsCommand(Command):
    """Print stats about FiftyOne datasets on disk.

    Examples::

        # Print stats about the given dataset on disk
        fiftyone datasets stats <name>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name",
            metavar="NAME",
            help="the name of the dataset",
        )
        parser.add_argument(
            "-m",
            "--include-media",
            action="store_true",
            help=(
                "whether to include stats about the size of the raw media in "
                "the dataset"
            ),
        )
        parser.add_argument(
            "-c",
            "--compressed",
            action="store_true",
            help=(
                "whether to return the sizes of collections in their "
                "compressed form on disk"
            ),
        )

    @staticmethod
    def execute(parser, args):
        dataset = fod.load_dataset(args.name)
        stats = dataset.stats(
            include_media=args.include_media, compressed=args.compressed
        )
        _print_dict_as_table(stats)


class DatasetsCreateCommand(Command):
    """Tools for creating FiftyOne datasets.

    Examples::

        # Create a dataset from the given data on disk
        fiftyone datasets create \\
            --name <name> --dataset-dir <dataset-dir> --type <type>

        # Create a dataset from a random subset of the data on disk
        fiftyone datasets create \\
            --name <name> --dataset-dir <dataset-dir> --type <type> \\
            --kwargs max_samples=50 shuffle=True

        # Create a dataset from the given samples JSON file
        fiftyone datasets create --json-path <json-path>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "-n",
            "--name",
            metavar="NAME",
            help="a name for the dataset",
        )
        parser.add_argument(
            "-d",
            "--dataset-dir",
            metavar="DATASET_DIR",
            help="the directory containing the dataset",
        )
        parser.add_argument(
            "-j",
            "--json-path",
            metavar="JSON_PATH",
            help="the path to a samples JSON file to load",
        )
        parser.add_argument(
            "-t",
            "--type",
            metavar="TYPE",
            help="the fiftyone.types.Dataset type of the dataset",
        )
        parser.add_argument(
            "-k",
            "--kwargs",
            nargs="+",
            metavar="KEY=VAL",
            action=_ParseKwargsAction,
            help=(
                "additional type-specific keyword arguments for "
                "`fiftyone.core.dataset.Dataset.from_dir()`"
            ),
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        dataset_dir = args.dataset_dir
        json_path = args.json_path
        dataset_type = etau.get_class(args.type) if args.type else None
        kwargs = args.kwargs or {}

        if json_path:
            dataset = fod.Dataset.from_json(json_path, name=name)
        else:
            dataset = fod.Dataset.from_dir(
                dataset_dir=dataset_dir,
                dataset_type=dataset_type,
                name=name,
                **kwargs,
            )

        dataset.persistent = True

        print("Dataset '%s' created" % dataset.name)


class DatasetsHeadCommand(Command):
    """Prints the first few samples in a FiftyOne dataset.

    Examples::

        # Print the first few samples in a dataset
        fiftyone datasets head <name>

        # Print the given number of samples from the head of a dataset
        fiftyone datasets head <name> --num-samples <num-samples>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the dataset"
        )
        parser.add_argument(
            "-n",
            "--num-samples",
            metavar="NUM_SAMPLES",
            type=int,
            default=3,
            help="the number of samples to print",
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        num_samples = args.num_samples

        dataset = fod.load_dataset(name)
        for sample in dataset.head(num_samples=num_samples):
            print(sample)


class DatasetsTailCommand(Command):
    """Prints the last few samples in a FiftyOne dataset.

    Examples::

        # Print the last few samples in a dataset
        fiftyone datasets tail <name>

        # Print the given number of samples from the tail of a dataset
        fiftyone datasets tail <name> --num-samples <num-samples>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the dataset"
        )
        parser.add_argument(
            "-n",
            "--num-samples",
            metavar="NUM_SAMPLES",
            type=int,
            default=3,
            help="the number of samples to print",
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        num_samples = args.num_samples

        dataset = fod.load_dataset(name)
        for sample in dataset.tail(num_samples=num_samples):
            print(sample)


class DatasetsStreamCommand(Command):
    """Stream samples in a FiftyOne dataset to the terminal.

    Examples::

        # Stream the samples of the dataset to the terminal
        fiftyone datasets stream <name>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the dataset"
        )

    @staticmethod
    def execute(parser, args):
        dataset = fod.load_dataset(args.name)
        fou.stream_objects(dataset)


class DatasetsExportCommand(Command):
    """Export FiftyOne datasets to disk in supported formats.

    Examples::

        # Export the dataset to disk in the specified format
        fiftyone datasets export <name> \\
            --export-dir <export-dir> --type <type> --label-field <label-field>

        # Export the dataset to disk in JSON format
        fiftyone datasets export <name> --json-path <json-path>

        # Perform a customized export of a dataset
        fiftyone datasets export <name> \\
            --type <type> \\
            --kwargs labels_path=/path/for/labels.json
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name",
            metavar="NAME",
            help="the name of the dataset to export",
        )
        parser.add_argument(
            "-d",
            "--export-dir",
            metavar="EXPORT_DIR",
            help="the directory in which to export the dataset",
        )
        parser.add_argument(
            "-j",
            "--json-path",
            metavar="JSON_PATH",
            help="the path to export the dataset in JSON format",
        )
        parser.add_argument(
            "-f",
            "--label-field",
            metavar="LABEL_FIELD",
            help="the name of the label field to export",
        )
        parser.add_argument(
            "-t",
            "--type",
            metavar="TYPE",
            help="the fiftyone.types.Dataset type in which to export",
        )
        parser.add_argument(
            "-k",
            "--kwargs",
            nargs="+",
            metavar="KEY=VAL",
            action=_ParseKwargsAction,
            help=(
                "additional type-specific keyword arguments for "
                "`fiftyone.core.collections.SampleCollection.export()`"
            ),
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        export_dir = args.export_dir
        json_path = args.json_path
        dataset_type = etau.get_class(args.type) if args.type else None
        label_field = args.label_field
        kwargs = args.kwargs or {}

        dataset = fod.load_dataset(name)

        if json_path:
            dataset.write_json(json_path)
            print("Dataset '%s' exported to '%s'" % (name, json_path))
        else:
            dataset.export(
                export_dir=export_dir,
                dataset_type=dataset_type,
                label_field=label_field,
                **kwargs,
            )

            if export_dir:
                print("Dataset '%s' exported to '%s'" % (name, export_dir))
            else:
                print("Dataset '%s' export complete" % name)


class DatasetsDrawCommand(Command):
    """Renders annotated versions of samples in FiftyOne datasets to disk.

    Examples::

        # Write annotated versions of the media in the dataset with the
        # specified label field(s) overlaid to disk
        fiftyone datasets draw <name> \\
            --output-dir <output-dir> --label-fields <list>,<of>,<fields>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name",
            metavar="NAME",
            help="the name of the dataset",
        )
        parser.add_argument(
            "-d",
            "--output-dir",
            metavar="OUTPUT_DIR",
            help="the directory to write the annotated media",
        )
        parser.add_argument(
            "-f",
            "--label-fields",
            metavar="LABEL_FIELDS",
            help="a comma-separated list of label fields to export",
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        output_dir = args.output_dir
        label_fields = args.label_fields

        dataset = fod.load_dataset(name)

        if label_fields is not None:
            label_fields = [f.strip() for f in label_fields.split(",")]

        dataset.draw_labels(output_dir, label_fields=label_fields)
        print("Rendered media written to '%s'" % output_dir)


class DatasetsRenameCommand(Command):
    """Rename FiftyOne datasets.

    Examples::

        # Rename the dataset
        fiftyone datasets rename <old-name> <new-name>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name",
            metavar="NAME",
            help="the name of the dataset",
        )
        parser.add_argument(
            "new_name",
            metavar="NEW_NAME",
            help="a new name for the dataset",
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        new_name = args.new_name

        dataset = fod.load_dataset(name)
        dataset.name = new_name
        print("Dataset '%s' renamed to '%s'" % (name, new_name))


class DatasetsDeleteCommand(Command):
    """Delete FiftyOne datasets.

    Examples::

        # Delete the datasets with the given name(s)
        fiftyone datasets delete <name1> <name2> ...

        # Delete the datasets whose names match the given glob pattern
        fiftyone datasets delete --glob-patt <glob-patt>

        # Delete all non-persistent datasets
        fiftyone datasets delete --non-persistent
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name",
            metavar="NAME",
            nargs="*",
            help="the dataset name(s) to delete",
        )
        parser.add_argument(
            "-g",
            "--glob-patt",
            metavar="GLOB_PATT",
            help="a glob pattern of datasets to delete",
        )
        parser.add_argument(
            "--non-persistent",
            action="store_true",
            help="delete all non-persistent datasets",
        )

    @staticmethod
    def execute(parser, args):
        for name in args.name:
            fod.delete_dataset(name, verbose=True)

        if args.glob_patt:
            fod.delete_datasets(args.glob_patt, verbose=True)

        if args.non_persistent:
            fod.delete_non_persistent_datasets(verbose=True)


class AnnotationCommand(Command):
    """Tools for working with the FiftyOne annotation API."""

    @staticmethod
    def setup(parser):
        subparsers = parser.add_subparsers(title="available commands")
        _register_command(subparsers, "config", AnnotationConfigCommand)

    @staticmethod
    def execute(parser, args):
        parser.print_help()


class AnnotationConfigCommand(Command):
    """Tools for working with your FiftyOne annotation config.

    Examples::

        # Print your entire annotation config
        fiftyone annotation config

        # Print a specific annotation config field
        fiftyone annotation config <field>

        # Print the location of your annotation config on disk (if one exists)
        fiftyone annotation config --locate
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "field",
            nargs="?",
            metavar="FIELD",
            help="an annotation config field to print",
        )
        parser.add_argument(
            "-l",
            "--locate",
            action="store_true",
            help="print the location of your annotation config on disk",
        )

    @staticmethod
    def execute(parser, args):
        if args.locate:
            annotation_config_path = focg.locate_annotation_config()
            if os.path.isfile(annotation_config_path):
                print(annotation_config_path)
            else:
                print(
                    "No annotation config file found at '%s'"
                    % annotation_config_path
                )

            return

        if args.field:
            field = getattr(fo.annotation_config, args.field)
            if etau.is_str(field):
                print(field)
            else:
                print(etas.json_to_str(field))
        else:
            print(fo.annotation_config)


class AppCommand(Command):
    """Tools for working with the FiftyOne App."""

    @staticmethod
    def setup(parser):
        subparsers = parser.add_subparsers(title="available commands")
        _register_command(subparsers, "config", AppConfigCommand)
        _register_command(subparsers, "launch", AppLaunchCommand)
        _register_command(subparsers, "view", AppViewCommand)
        _register_command(subparsers, "connect", AppConnectCommand)

    @staticmethod
    def execute(parser, args):
        parser.print_help()


class AppConfigCommand(Command):
    """Tools for working with your FiftyOne App config.

    Examples::

        # Print your entire App config
        fiftyone app config

        # Print a specific App config field
        fiftyone app config <field>

        # Print the location of your App config on disk (if one exists)
        fiftyone app config --locate
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "field",
            nargs="?",
            metavar="FIELD",
            help="an App config field to print",
        )
        parser.add_argument(
            "-l",
            "--locate",
            action="store_true",
            help="print the location of your App config on disk",
        )

    @staticmethod
    def execute(parser, args):
        if args.locate:
            app_config_path = focg.locate_app_config()
            if os.path.isfile(app_config_path):
                print(app_config_path)
            else:
                print("No App config file found at '%s'" % app_config_path)

            return

        if args.field:
            field = getattr(fo.app_config, args.field)
            if etau.is_str(field):
                print(field)
            else:
                print(etas.json_to_str(field))
        else:
            print(fo.app_config)


class AppLaunchCommand(Command):
    """Launch the FiftyOne App.

    Examples::

        # Launch the App
        fiftyone app launch

        # Launch the App with the given dataset loaded
        fiftyone app launch <name>

        # Launch a remote App session
        fiftyone app launch ... --remote

        # Launch a desktop App session
        fiftyone app launch ... --desktop
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name",
            metavar="NAME",
            nargs="?",
            help="the name of a dataset to open",
        )
        parser.add_argument(
            "-p",
            "--port",
            metavar="PORT",
            default=None,
            type=int,
            help="the port number to use",
        )
        parser.add_argument(
            "-A",
            "--address",
            metavar="ADDRESS",
            default=None,
            type=str,
            help="the address (server name) to use",
        )
        parser.add_argument(
            "-r",
            "--remote",
            action="store_true",
            help="whether to launch a remote App session",
        )
        parser.add_argument(
            "-a",
            "--desktop",
            action="store_true",
            help="whether to launch a desktop App instance",
        )
        parser.add_argument(
            "-w",
            "--wait",
            metavar="WAIT",
            default=3,
            type=float,
            help=(
                "the number of seconds to wait for a new App connection "
                "before returning if all connections are lost. If negative, "
                "the process will wait forever, regardless of connections"
            ),
        )

    @staticmethod
    def execute(parser, args):
        # If desktop wasn't explicitly requested, fallback to default
        desktop = args.desktop or None

        if args.name:
            dataset = fod.load_dataset(args.name)
        else:
            dataset = None

        session = fos.launch_app(
            dataset=dataset,
            port=args.port,
            address=args.address,
            remote=args.remote,
            desktop=desktop,
        )

        _watch_session(session, args.wait)


def _watch_session(session, wait):
    # Automated tests may set `FIFTYONE_EXIT` so they can immediately exit
    if os.environ.get("FIFTYONE_EXIT", False):
        return

    try:
        print("\nTo exit, close the App or press ctrl + c\n")
        session.wait(wait)
    except KeyboardInterrupt:
        pass


def _wait():
    print("\nTo exit, press ctrl + c\n")

    try:
        while True:
            time.sleep(10)
    except KeyboardInterrupt:
        pass


class AppViewCommand(Command):
    """View datasets in the App without persisting them to the database.

    Examples::

        # View a dataset stored on disk in the App
        fiftyone app view --dataset-dir <dataset-dir> --type <type>

        # View a zoo dataset in the App
        fiftyone app view --zoo-dataset <name> --splits <split1> ...

        # View a directory of images in the App
        fiftyone app view --images-dir <images-dir>

        # View a glob pattern of images in the App
        fiftyone app view --images-patt <images-patt>

        # View a directory of videos in the App
        fiftyone app view --videos-dir <videos-dir>

        # View a glob pattern of videos in the App
        fiftyone app view --videos-patt <videos-patt>

        # View a dataset stored in JSON format on disk in the App
        fiftyone app view --json-path <json-path>

        # View the dataset in a remote App session
        fiftyone app view ... --remote

        # View the dataset using the desktop App
        fiftyone app view ... --desktop

        # View a random subset of the data stored on disk in the App
        fiftyone app view ... --kwargs max_samples=50 shuffle=True
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "-n", "--name", metavar="NAME", help="a name for the dataset"
        )
        parser.add_argument(
            "-d",
            "--dataset-dir",
            metavar="DATASET_DIR",
            help="the directory containing the dataset to view",
        )
        parser.add_argument(
            "-t",
            "--type",
            metavar="TYPE",
            help="the fiftyone.types.Dataset type of the dataset",
        )
        parser.add_argument(
            "-z",
            "--zoo-dataset",
            metavar="NAME",
            help="the name of a zoo dataset to view",
        )
        parser.add_argument(
            "-s",
            "--splits",
            metavar="SPLITS",
            nargs="+",
            help="the dataset splits to load",
        )
        parser.add_argument(
            "--images-dir",
            metavar="IMAGES_DIR",
            help="the path to a directory of images",
        )
        parser.add_argument(
            "--images-patt",
            metavar="IMAGES_PATT",
            help="a glob pattern of images",
        )
        parser.add_argument(
            "--videos-dir",
            metavar="VIDEOS_DIR",
            help="the path to a directory of videos",
        )
        parser.add_argument(
            "--videos-patt",
            metavar="VIDEOS_PATT",
            help="a glob pattern of videos",
        )
        parser.add_argument(
            "-j",
            "--json-path",
            metavar="JSON_PATH",
            help="the path to a samples JSON file to view",
        )
        parser.add_argument(
            "-p",
            "--port",
            metavar="PORT",
            default=None,
            type=int,
            help="the port number to use",
        )
        parser.add_argument(
            "-A",
            "--address",
            metavar="ADDRESS",
            default=None,
            type=str,
            help="the address (server name) to use",
        )
        parser.add_argument(
            "-r",
            "--remote",
            action="store_true",
            help="whether to launch a remote App session",
        )
        parser.add_argument(
            "-a",
            "--desktop",
            action="store_true",
            help="whether to launch a desktop App instance",
        )
        parser.add_argument(
            "-w",
            "--wait",
            metavar="WAIT",
            default=3,
            type=float,
            help=(
                "the number of seconds to wait for a new App connection "
                "before returning if all connections are lost. If negative, "
                "the process will wait forever, regardless of connections"
            ),
        )
        parser.add_argument(
            "-k",
            "--kwargs",
            nargs="+",
            metavar="KEY=VAL",
            action=_ParseKwargsAction,
            help=(
                "additional type-specific keyword arguments for "
                "`fiftyone.core.dataset.Dataset.from_dir()`"
            ),
        )

    @staticmethod
    def execute(parser, args):
        if args.zoo_dataset:
            # View a zoo dataset
            name = args.zoo_dataset
            splits = args.splits
            dataset_dir = args.dataset_dir
            kwargs = args.kwargs or {}

            dataset = fozd.load_zoo_dataset(
                name, splits=splits, dataset_dir=dataset_dir, **kwargs
            )
        elif args.images_dir:
            # View a directory of images
            name = args.name
            images_dir = args.images_dir
            dataset = fod.Dataset.from_images_dir(images_dir, name=name)
        elif args.images_patt:
            # View a glob pattern of images
            name = args.name
            images_patt = args.images_patt
            dataset = fod.Dataset.from_images_patt(images_patt, name=name)
        elif args.videos_dir:
            # View a directory of images
            name = args.name
            videos_dir = args.videos_dir
            dataset = fod.Dataset.from_videos_dir(videos_dir, name=name)
        elif args.videos_patt:
            # View a glob pattern of videos
            name = args.name
            videos_patt = args.videos_patt
            dataset = fod.Dataset.from_videos_patt(videos_patt, name=name)
        elif args.json_path:
            # View a dataset from a JSON file
            name = args.name
            json_path = args.json_path
            dataset = fod.Dataset.from_json(json_path, name=name)
        else:
            # View a dataset from disk
            name = args.name
            dataset_dir = args.dataset_dir
            dataset_type = etau.get_class(args.type)
            kwargs = args.kwargs or {}

            dataset = fod.Dataset.from_dir(
                dataset_dir=dataset_dir,
                dataset_type=dataset_type,
                name=name,
                **kwargs,
            )

        # If desktop wasn't explicitly requested, fallback to default
        desktop = args.desktop or None

        session = fos.launch_app(
            dataset=dataset,
            port=args.port,
            address=args.address,
            remote=args.remote,
            desktop=desktop,
        )

        _watch_session(session, args.wait)


class AppConnectCommand(Command):
    """Connect to a remote FiftyOne App in your web browser.

    Examples::

        # Connect to a remote App with port forwarding already configured
        fiftyone app connect

        # Connect to a remote App session
        fiftyone app connect --destination <destination> --port <port>

        # Connect to a remote App session using an ssh key
        fiftyone app connect ... --ssh-key <path/to/key>

        # Connect to a remote App using a custom local port
        fiftyone app connect ... --local-port <port>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "-d",
            "--destination",
            metavar="DESTINATION",
            type=str,
            help="the destination to connect to, e.g., [username@]hostname",
        )
        parser.add_argument(
            "-p",
            "--port",
            metavar="PORT",
            default=None,
            type=int,
            help="the remote port to connect to",
        )
        parser.add_argument(
            "-l",
            "--local-port",
            metavar="PORT",
            default=None,
            type=int,
            help="the local port to use to serve the App",
        )
        parser.add_argument(
            "-i",
            "--ssh-key",
            metavar="KEY",
            default=None,
            type=str,
            help="optional ssh key to use to login",
        )

    @staticmethod
    def execute(parser, args):
        remote_port = args.port or fo.config.default_app_port
        local_port = args.local_port or fo.config.default_app_port

        if args.destination:
            if sys.platform.startswith("win"):
                raise RuntimeError(
                    "This command is currently not supported on Windows"
                )

            control_path = os.path.join(
                foc.FIFTYONE_CONFIG_DIR, "tmp", "ssh.sock"
            )
            etau.ensure_basedir(control_path)

            ssh_call = [
                "ssh",
                "-f",
                "-N",
                "-M",
                "-S",
                control_path,
                "-L",
                "%d:127.0.0.1:%d" % (local_port, remote_port),
            ]

            if args.ssh_key:
                ssh_call += ["-i", args.ssh_key]

            ssh_call.append(args.destination)

            # Port forwarding
            ret = subprocess.call(ssh_call)
            if ret != 0:
                print("ssh failed with exit code %r" % ret)
                return

            def stop_port_forward():
                subprocess.call(
                    [
                        "ssh",
                        "-S",
                        control_path,
                        "-O",
                        "exit",
                        args.destination,
                    ],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                )

            fou.call_on_exit(stop_port_forward)

        url = "http://localhost:%d/" % local_port
        webbrowser.open(url, new=2)

        _wait()


class ZooCommand(Command):
    """Tools for working with the FiftyOne Zoo."""

    @staticmethod
    def setup(parser):
        subparsers = parser.add_subparsers(title="available commands")
        _register_command(subparsers, "datasets", DatasetZooCommand)
        _register_command(subparsers, "models", ModelZooCommand)

    @staticmethod
    def execute(parser, args):
        parser.print_help()


class DatasetZooCommand(Command):
    """Tools for working with the FiftyOne Dataset Zoo."""

    @staticmethod
    def setup(parser):
        subparsers = parser.add_subparsers(title="available commands")
        _register_command(subparsers, "list", DatasetZooListCommand)
        _register_command(subparsers, "find", DatasetZooFindCommand)
        _register_command(subparsers, "info", DatasetZooInfoCommand)
        _register_command(subparsers, "download", DatasetZooDownloadCommand)
        _register_command(subparsers, "load", DatasetZooLoadCommand)
        _register_command(subparsers, "delete", DatasetZooDeleteCommand)

    @staticmethod
    def execute(parser, args):
        parser.print_help()


class DatasetZooListCommand(Command):
    """List datasets in the FiftyOne Dataset Zoo.

    Examples::

        # List available datasets
        fiftyone zoo datasets list

        # List available datasets (names only)
        fiftyone zoo datasets list --names-only

        # List downloaded datasets
        fiftyone zoo datasets list --downloaded-only

        # List available datasets from the given source
        fiftyone zoo datasets list --source <source>

        # List available datasets with the given tag
        fiftyone zoo datasets list --tags <tag>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "-n",
            "--names-only",
            action="store_true",
            help="only show dataset names",
        )
        parser.add_argument(
            "-d",
            "--downloaded-only",
            action="store_true",
            help="only show datasets that have been downloaded",
        )
        parser.add_argument(
            "-s",
            "--source",
            metavar="SOURCE",
            help="only show datasets available from the specified source",
        )
        parser.add_argument(
            "-t",
            "--tags",
            metavar="TAGS",
            help="only show datasets with the specified tag or list,of,tags",
        )
        parser.add_argument(
            "-b",
            "--base-dir",
            metavar="BASE_DIR",
            help=(
                "a custom base directory in which to search for downloaded "
                "datasets"
            ),
        )

    @staticmethod
    def execute(parser, args):
        names_only = args.names_only
        downloaded_only = args.downloaded_only
        match_source = args.source
        match_tags = args.tags

        all_datasets = fozd._get_zoo_datasets()
        all_sources, default_source = fozd._get_zoo_dataset_sources()

        base_dir = args.base_dir
        downloaded_datasets = fozd.list_downloaded_zoo_datasets(
            base_dir=base_dir
        )

        _print_zoo_dataset_list(
            downloaded_datasets,
            all_datasets,
            all_sources,
            default_source,
            downloaded_only=downloaded_only,
            match_tags=match_tags,
            match_source=match_source,
            names_only=names_only,
        )


def _print_zoo_dataset_list(
    downloaded_datasets,
    all_datasets,
    all_sources,
    default_source,
    downloaded_only=False,
    match_tags=None,
    match_source=None,
    names_only=False,
):
    if match_tags is not None:
        match_tags = match_tags.split(",")

    available_datasets = defaultdict(dict)
    for source, datasets in all_datasets.items():
        for name, zoo_dataset_cls in datasets.items():
            available_datasets[name][source] = zoo_dataset_cls()

    records = []

    # Iterate over available datasets
    for name in sorted(available_datasets):
        if downloaded_only and name not in downloaded_datasets:
            continue

        dataset_sources = available_datasets[name]

        if match_source is not None and match_source not in dataset_sources:
            continue

        tags = None
        for source, zoo_model in dataset_sources.items():
            if tags is None or source == default_source:
                tags = zoo_model.tags

        if (match_tags is not None) and (
            tags is None or not all(tag in tags for tag in match_tags)
        ):
            continue

        # Check for downloaded splits
        if name in downloaded_datasets:
            dataset_dir, info = downloaded_datasets[name]
        else:
            dataset_dir, info = None, None

        if names_only:
            records.append(name)
            continue

        # Get available splits across all sources
        splits = set()
        for zoo_dataset in dataset_sources.values():
            if zoo_dataset.has_splits:
                splits.update(zoo_dataset.supported_splits)
            else:
                splits.add("")

        # Iterate over available splits
        for split in sorted(splits):
            # Get available sources for the split
            srcs = []
            for source in all_sources:
                if source not in dataset_sources:
                    srcs.append("")
                    continue

                zoo_dataset = dataset_sources[source]
                if split and zoo_dataset.has_split(split):
                    srcs.append("\u2713")
                elif not split and not zoo_dataset.has_splits:
                    srcs.append("\u2713")
                else:
                    srcs.append("")

            # Get split directory
            if not split and dataset_dir:
                split_dir = dataset_dir
            elif split and info and info.is_split_downloaded(split):
                split_dir = zoo_dataset.get_split_dir(dataset_dir, split)
            else:
                split_dir = ""

            if downloaded_only and not split_dir:
                continue

            tags_str = ",".join(tags) if tags else ""
            is_downloaded = "\u2713" if split_dir else ""

            records.append(
                (name, tags_str, split, is_downloaded, split_dir) + tuple(srcs)
            )

    if names_only:
        for name in records:
            print(name)

        return

    headers = ["name", "tags", "split", "downloaded", "dataset_dir"]
    for source in all_sources:
        if source == default_source:
            source += " (*)"

        headers.append(source)

    table_str = tabulate(records, headers=headers, tablefmt=_TABLE_FORMAT)
    print(table_str)


class DatasetZooFindCommand(Command):
    """Locate the downloaded zoo dataset on disk.

    Examples::

        # Print the location of the downloaded zoo dataset on disk
        fiftyone zoo datasets find <name>

        # Print the location of a specific split of the dataset
        fiftyone zoo datasets find <name> --split <split>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the dataset"
        )
        parser.add_argument(
            "-s",
            "--split",
            metavar="SPLIT",
            help="a dataset split",
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        split = args.split

        dataset_dir = fozd.find_zoo_dataset(name, split=split)
        print(dataset_dir)


class DatasetZooInfoCommand(Command):
    """Print information about datasets in the FiftyOne Dataset Zoo.

    Examples::

        # Print information about a zoo dataset
        fiftyone zoo datasets info <name>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the dataset"
        )
        parser.add_argument(
            "-b",
            "--base-dir",
            metavar="BASE_DIR",
            help=(
                "a custom base directory in which to search for downloaded "
                "datasets"
            ),
        )

    @staticmethod
    def execute(parser, args):
        name = args.name

        # Print dataset info
        zoo_dataset = fozd.get_zoo_dataset(name)
        print(
            "***** Dataset description *****\n%s"
            % textwrap.dedent("    " + zoo_dataset.__doc__)
        )

        # Check if dataset is downloaded
        base_dir = args.base_dir
        downloaded_datasets = fozd.list_downloaded_zoo_datasets(
            base_dir=base_dir
        )

        if zoo_dataset.has_tags:
            print("***** Tags *****")
            print("%s\n" % ", ".join(zoo_dataset.tags))

        if zoo_dataset.has_splits:
            print("***** Supported splits *****")
            print("%s\n" % ", ".join(zoo_dataset.supported_splits))

        print("***** Dataset location *****")
        if name not in downloaded_datasets:
            print("Dataset '%s' is not downloaded" % name)
        else:
            dataset_dir, info = downloaded_datasets[name]
            print(dataset_dir)
            print("\n***** Dataset info *****")
            print(info)


class DatasetZooDownloadCommand(Command):
    """Download zoo datasets.

    Examples::

        # Download the entire zoo dataset
        fiftyone zoo datasets download <name>

        # Download the specified split(s) of the zoo dataset
        fiftyone zoo datasets download <name> --splits <split1> ...

        # Download the zoo dataset to a custom directory
        fiftyone zoo datasets download <name> --dataset-dir <dataset-dir>

        # Download a zoo dataset that requires extra keyword arguments
        fiftyone zoo datasets download <name> \\
            --kwargs source_dir=/path/to/source/files
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the dataset"
        )
        parser.add_argument(
            "-s",
            "--splits",
            metavar="SPLITS",
            nargs="+",
            help="the dataset splits to download",
        )
        parser.add_argument(
            "-d",
            "--dataset-dir",
            metavar="DATASET_DIR",
            help="a custom directory to which to download the dataset",
        )
        parser.add_argument(
            "-k",
            "--kwargs",
            nargs="+",
            metavar="KEY=VAL",
            action=_ParseKwargsAction,
            help=(
                "optional dataset-specific keyword arguments for "
                "`fiftyone.zoo.download_zoo_dataset()`"
            ),
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        splits = args.splits
        dataset_dir = args.dataset_dir
        kwargs = args.kwargs or {}

        fozd.download_zoo_dataset(
            name, splits=splits, dataset_dir=dataset_dir, **kwargs
        )


class DatasetZooLoadCommand(Command):
    """Load zoo datasets as persistent FiftyOne datasets.

    Examples::

        # Load the zoo dataset with the given name
        fiftyone zoo datasets load <name>

        # Load the specified split(s) of the zoo dataset
        fiftyone zoo datasets load <name> --splits <split1> ...

        # Load the zoo dataset with a custom name
        fiftyone zoo datasets load <name> --dataset-name <dataset-name>

        # Load the zoo dataset from a custom directory
        fiftyone zoo datasets load <name> --dataset-dir <dataset-dir>

        # Load a zoo dataset that requires custom keyword arguments
        fiftyone zoo datasets load <name> \\
            --kwargs source_dir=/path/to/source_files

        # Load a random subset of a zoo dataset
        fiftyone zoo datasets load <name> \\
            --kwargs max_samples=50 shuffle=True
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the dataset"
        )
        parser.add_argument(
            "-s",
            "--splits",
            metavar="SPLITS",
            nargs="+",
            help="the dataset splits to load",
        )
        parser.add_argument(
            "-n",
            "--dataset-name",
            metavar="DATASET_NAME",
            help="a custom name to give the FiftyOne dataset",
        )
        parser.add_argument(
            "-d",
            "--dataset-dir",
            metavar="DATASET_DIR",
            help="a custom directory in which the dataset is downloaded",
        )
        parser.add_argument(
            "-k",
            "--kwargs",
            nargs="+",
            metavar="KEY=VAL",
            action=_ParseKwargsAction,
            help=(
                "additional dataset-specific keyword arguments for "
                "`fiftyone.zoo.load_zoo_dataset()`"
            ),
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        splits = args.splits
        dataset_name = args.dataset_name
        dataset_dir = args.dataset_dir
        kwargs = args.kwargs or {}

        dataset = fozd.load_zoo_dataset(
            name,
            splits=splits,
            dataset_name=dataset_name,
            dataset_dir=dataset_dir,
            **kwargs,
        )
        dataset.persistent = True


class DatasetZooDeleteCommand(Command):
    """Deletes the local copy of the zoo dataset on disk.

    Examples::

        # Delete an entire zoo dataset from disk
        fiftyone zoo datasets delete <name>

        # Delete a specific split of a zoo dataset from disk
        fiftyone zoo datasets delete <name> --split <split>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the dataset"
        )
        parser.add_argument(
            "-s",
            "--split",
            metavar="SPLIT",
            help="a dataset split",
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        split = args.split
        fozd.delete_zoo_dataset(name, split=split)


class ModelZooCommand(Command):
    """Tools for working with the FiftyOne Model Zoo."""

    @staticmethod
    def setup(parser):
        subparsers = parser.add_subparsers(title="available commands")
        _register_command(subparsers, "list", ModelZooListCommand)
        _register_command(subparsers, "find", ModelZooFindCommand)
        _register_command(subparsers, "info", ModelZooInfoCommand)
        _register_command(
            subparsers, "requirements", ModelZooRequirementsCommand
        )
        _register_command(subparsers, "download", ModelZooDownloadCommand)
        _register_command(subparsers, "apply", ModelZooApplyCommand)
        _register_command(subparsers, "embed", ModelZooEmbedCommand)
        _register_command(subparsers, "delete", ModelZooDeleteCommand)

    @staticmethod
    def execute(parser, args):
        parser.print_help()


class ModelZooListCommand(Command):
    """List datasets in the FiftyOne Model Zoo.

    Examples::

        # List available models
        fiftyone zoo models list

        # List available models (names only)
        fiftyone zoo models list --names-only

        # List downloaded models
        fiftyone zoo models list --downloaded-only

        # List available models with the given tag
        fiftyone zoo models list --tags <tag>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "-n",
            "--names-only",
            action="store_true",
            help="only show model names",
        )
        parser.add_argument(
            "-d",
            "--downloaded-only",
            action="store_true",
            help="only show models that have been downloaded",
        )
        parser.add_argument(
            "-t",
            "--tags",
            metavar="TAGS",
            help="only show models with the specified tag or list,of,tags",
        )

    @staticmethod
    def execute(parser, args):
        names_only = args.names_only
        downloaded_only = args.downloaded_only
        match_tags = args.tags

        models_manifest = fozm._load_zoo_models_manifest()
        downloaded_models = fozm.list_downloaded_zoo_models()

        _print_zoo_models_list(
            models_manifest,
            downloaded_models,
            downloaded_only=downloaded_only,
            match_tags=match_tags,
            names_only=names_only,
        )


def _print_zoo_models_list(
    models_manifest,
    downloaded_models,
    downloaded_only=False,
    match_tags=None,
    names_only=False,
):
    if match_tags is not None:
        match_tags = match_tags.split(",")

    records = []
    for model in sorted(models_manifest.models, key=lambda model: model.name):
        name = model.name

        if downloaded_only and name not in downloaded_models:
            continue

        if (match_tags is not None) and not all(
            model.has_tag(tag) for tag in match_tags
        ):
            continue

        if names_only:
            records.append(name)
            continue

        if name in downloaded_models:
            is_downloaded = "\u2713"
            model_path = downloaded_models[name][0]
        else:
            is_downloaded = ""
            model_path = ""

        tags = ",".join(model.tags or [])

        records.append((name, tags, is_downloaded, model_path))

    if names_only:
        for name in records:
            print(name)

        return

    headers = ["name", "tags", "downloaded", "model_path"]
    table_str = tabulate(records, headers=headers, tablefmt=_TABLE_FORMAT)
    print(table_str)


def _print_zoo_models_list_sample(models_manifest, downloaded_models):
    all_models = [model.name for model in models_manifest]

    records = []
    for name in sorted(all_models):
        if name in downloaded_models:
            is_downloaded = "\u2713"
            model_path = downloaded_models[name][0]
        else:
            is_downloaded = ""
            model_path = ""

        records.append((name, is_downloaded, model_path))

    headers = ["name", "downloaded", "model_path"]
    table_str = tabulate(records, headers=headers, tablefmt=_TABLE_FORMAT)
    print(table_str)


class ModelZooFindCommand(Command):
    """Locate the downloaded zoo model on disk.

    Examples::

        # Print the location of the downloaded zoo model on disk
        fiftyone zoo models find <name>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the model"
        )

    @staticmethod
    def execute(parser, args):
        name = args.name

        model_path = fozm.find_zoo_model(name)
        print(model_path)


class ModelZooInfoCommand(Command):
    """Print information about models in the FiftyOne Model Zoo.

    Examples::

        # Print information about a zoo model
        fiftyone zoo models info <name>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the model"
        )

    @staticmethod
    def execute(parser, args):
        name = args.name

        # Print model info
        zoo_model = fozm.get_zoo_model(name)
        print("***** Model description *****\n%s\n" % str(zoo_model))

        # Check if model is downloaded
        print("***** Model location *****")
        if not fozm.is_zoo_model_downloaded(name):
            print("Model '%s' is not downloaded" % name)
        else:
            model_path = fozm.find_zoo_model(name)
            print(model_path)


class ModelZooRequirementsCommand(Command):
    """Handles package requirements for zoo models.

    Examples::

        # Print requirements for a zoo model
        fiftyone zoo models requirements <name> --print

        # Install any requirements for the zoo model
        fiftyone zoo models requirements <name> --install

        # Ensures that the requirements for the zoo model are satisfied
        fiftyone zoo models requirements <name> --ensure
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the model"
        )
        parser.add_argument(
            "-p",
            "--print",
            action="store_true",
            help="print the requirements for the zoo model",
        )
        parser.add_argument(
            "-i",
            "--install",
            action="store_true",
            help="install any requirements for the zoo model",
        )
        parser.add_argument(
            "-e",
            "--ensure",
            action="store_true",
            help="ensure the requirements for the zoo model are satisfied",
        )
        parser.add_argument(
            "--error-level",
            metavar="LEVEL",
            type=int,
            help=(
                "the error level (0=error, 1=warn, 2=ignore) to use when "
                "installing or ensuring model requirements"
            ),
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        error_level = args.error_level

        if args.print or (not args.install and not args.ensure):
            zoo_model = fozm.get_zoo_model(name)
            _print_model_requirements(zoo_model)

        if args.install:
            fozm.install_zoo_model_requirements(name, error_level=error_level)

        if args.ensure:
            fozm.ensure_zoo_model_requirements(name, error_level=error_level)


def _print_model_requirements(zoo_model):
    requirements = zoo_model.requirements
    if requirements is None:
        return

    # Model requirements
    print("***** Model requirements *****")
    print(requirements)

    # Current machine specs
    print("\n***** Current machine *****")
    if etau.has_gpu():
        print("GPU: yes")
        print("CUDA version: %s" % etau.get_cuda_version())
        print("cuDNN version: %s" % etau.get_cudnn_version())
    else:
        print("GPU: no")


class ModelZooDownloadCommand(Command):
    """Download zoo models.

    Examples::

        # Download the zoo model
        fiftyone zoo models download <name>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the zoo model"
        )
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help=(
                "whether to force download the model if it is already "
                "downloaded"
            ),
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        force = args.force
        fozm.download_zoo_model(name, overwrite=force)


class ModelZooApplyCommand(Command):
    """Apply zoo models to datasets.

    Examples::

        # Apply the zoo model to the dataset
        fiftyone zoo models apply <model-name> <dataset-name> <label-field>

        # Apply a zoo classifier with some customized parameters
        fiftyone zoo models apply \\
            <model-name> <dataset-name> <label-field> \\
            --confidence-thresh 0.7 \\
            --store-logits \\
            --batch-size 32
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "model_name",
            metavar="MODEL_NAME",
            help="the name of the zoo model",
        )
        parser.add_argument(
            "dataset_name",
            metavar="DATASET_NAME",
            help="the name of the FiftyOne dataset to process",
        )
        parser.add_argument(
            "label_field",
            metavar="LABEL_FIELD",
            help="the name of the field in which to store the predictions",
        )
        parser.add_argument(
            "-b",
            "--batch-size",
            metavar="BATCH_SIZE",
            default=None,
            type=int,
            help="an optional batch size to use during inference",
        )
        parser.add_argument(
            "-t",
            "--confidence-thresh",
            metavar="THRESH",
            default=None,
            type=float,
            help=(
                "an optional confidence threshold to apply to any applicable "
                "labels generated by the model"
            ),
        )
        parser.add_argument(
            "-l",
            "--store-logits",
            action="store_true",
            help="store logits for the predictions",
        )
        parser.add_argument(
            "-i",
            "--install",
            action="store_true",
            help="install any requirements for the zoo model",
        )
        parser.add_argument(
            "--error-level",
            metavar="LEVEL",
            type=int,
            help=(
                "the error level (0=error, 1=warn, 2=ignore) to use when "
                "installing or ensuring model requirements"
            ),
        )

    @staticmethod
    def execute(parser, args):
        model = fozm.load_zoo_model(
            args.model_name,
            install_requirements=args.install,
            error_level=args.error_level,
        )

        dataset = fod.load_dataset(args.dataset_name)

        dataset.apply_model(
            model,
            args.label_field,
            confidence_thresh=args.confidence_thresh,
            store_logits=args.store_logits,
            batch_size=args.batch_size,
        )


class ModelZooEmbedCommand(Command):
    """Generate embeddings for datasets with zoo models.

    Examples::

        # Generate embeddings for the dataset with the zoo model
        fiftyone zoo models embed <model-name> <dataset-name> <embeddings-field>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "model_name",
            metavar="MODEL_NAME",
            help="the name of the zoo model",
        )
        parser.add_argument(
            "dataset_name",
            metavar="DATASET_NAME",
            help="the name of the FiftyOne dataset to process",
        )
        parser.add_argument(
            "embeddings_field",
            metavar="EMBEDDINGS_FIELD",
            help="the name of the field in which to store the embeddings",
        )
        parser.add_argument(
            "-b",
            "--batch-size",
            metavar="BATCH_SIZE",
            default=None,
            type=int,
            help="an optional batch size to use during inference",
        )
        parser.add_argument(
            "-i",
            "--install",
            action="store_true",
            help="install any requirements for the zoo model",
        )
        parser.add_argument(
            "--error-level",
            metavar="LEVEL",
            type=int,
            help=(
                "the error level (0=error, 1=warn, 2=ignore) to use when "
                "installing or ensuring model requirements"
            ),
        )

    @staticmethod
    def execute(parser, args):
        model = fozm.load_zoo_model(
            args.model_name,
            install_requirements=args.install,
            error_level=args.error_level,
        )
        dataset = fod.load_dataset(args.dataset_name)
        dataset.compute_embeddings(
            model,
            embeddings_field=args.embeddings_field,
            batch_size=args.batch_size,
        )


class ModelZooDeleteCommand(Command):
    """Deletes the local copy of the zoo model on disk.

    Examples::

        # Delete the zoo model from disk
        fiftyone zoo models delete <name>
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="NAME", help="the name of the model"
        )

    @staticmethod
    def execute(parser, args):
        name = args.name
        fozm.delete_zoo_model(name)


class MigrateCommand(Command):
    """Tools for migrating the FiftyOne database.

    See :ref:`this page <database-migrations>` for more information about
    migrating FiftyOne deployments.

    Examples::

        # Print information about the current revisions of all datasets
        fiftyone migrate --info

        # Migrate the database and all datasets to the current client version
        fiftyone migrate --all

        # Migrate to a specific revision
        fiftyone migrate --all --version <VERSION>

        # Migrate a specific dataset
        fiftyone migrate ... --dataset-name <DATASET_NAME>

        # Update the database version without migrating any existing datasets
        fiftyone migrate
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "-i",
            "--info",
            action="store_true",
            help="whether to print info about the current revisions",
        )
        parser.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="whether to migrate the database and all datasets",
        )
        parser.add_argument(
            "-v",
            "--version",
            metavar="VERSION",
            help="the revision to migrate to",
        )
        parser.add_argument(
            "-n",
            "--dataset-name",
            nargs="+",
            metavar="DATASET_NAME",
            help="the name of a specific dataset to migrate",
        )
        parser.add_argument(
            "--error-level",
            metavar="LEVEL",
            type=int,
            default=1,
            help=(
                "the error level (0=error, 1=warn, 2=ignore) to use when "
                "migrating individual datasets"
            ),
        )
        parser.add_argument(
            "--verbose",
            action="store_true",
            help="whether to log incremental migrations that are performed",
        )

    @staticmethod
    def execute(parser, args):
        if args.info:
            db_ver = fom.get_database_revision() or ""

            if args.dataset_name is not None:
                for name in args.dataset_name:
                    print(fom.get_dataset_revision(name))

                return

            dataset_vers = {
                name: fom.get_dataset_revision(name)
                for name in fod.list_datasets()
            }

            _print_migration_table(db_ver, dataset_vers)
            return

        if args.all:
            fom.migrate_all(
                destination=args.version,
                error_level=args.error_level,
                verbose=args.verbose,
            )
            return

        if args.dataset_name:
            for name in args.dataset_name:
                fom.migrate_dataset_if_necessary(
                    name,
                    destination=args.version,
                    error_level=args.error_level,
                    verbose=args.verbose,
                )
        else:
            fom.migrate_database_if_necessary(
                destination=args.version,
                verbose=args.verbose,
            )


def _print_migration_table(db_ver, dataset_vers):
    print("Client version: %s" % foc.VERSION)

    if foc.COMPATIBLE_VERSIONS:
        print("Compatible versions: %s" % foc.COMPATIBLE_VERSIONS)
        print("")

    print("Database version: %s" % db_ver)

    if dataset_vers:
        print("")
        _print_dict_as_table(dataset_vers, headers=["dataset", "version"])


class UtilsCommand(Command):
    """FiftyOne utilities."""

    @staticmethod
    def setup(parser):
        subparsers = parser.add_subparsers(title="available commands")
        _register_command(
            subparsers, "compute-metadata", ComputeMetadataCommand
        )
        _register_command(
            subparsers, "transform-images", TransformImagesCommand
        )
        _register_command(
            subparsers, "transform-videos", TransformVideosCommand
        )

    @staticmethod
    def execute(parser, args):
        parser.print_help()


class ComputeMetadataCommand(Command):
    """Populates the `metadata` field of all samples in the dataset.

    Examples::

        # Populate all missing `metadata` sample fields
        fiftyone utils compute-metadata <dataset-name>

        # (Re)-populate the `metadata` field for all samples
        fiftyone utils compute-metadata <dataset-name> --overwrite
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="DATASET_NAME", help="the name of the dataset"
        )
        parser.add_argument(
            "-o",
            "--overwrite",
            action="store_true",
            help="whether to overwrite existing metadata",
        )
        parser.add_argument(
            "-n",
            "--num-workers",
            default=None,
            type=int,
            help=(
                "the number of worker processes to use. The default is "
                "`multiprocessing.cpu_count()`"
            ),
        )
        parser.add_argument(
            "-s",
            "--skip-failures",
            action="store_true",
            help=(
                "whether to gracefully continue without raising an error if "
                "metadata cannot be computed for a sample"
            ),
        )

    @staticmethod
    def execute(parser, args):
        dataset = fod.load_dataset(args.name)
        dataset.compute_metadata(
            overwrite=args.overwrite,
            num_workers=args.num_workers,
            skip_failures=args.skip_failures,
        )


class TransformImagesCommand(Command):
    """Transforms the images in a dataset per the specified parameters.

    Examples::

        # Convert the images in the dataset to PNGs
        fiftyone utils transform-images <dataset-name> --ext .png --delete-originals

        # Ensure that no images in the dataset exceed 1920 x 1080
        fiftyone utils transform-images <dataset-name> --max-size 1920,1080
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="DATASET_NAME", help="the name of the dataset"
        )
        parser.add_argument(
            "--size",
            metavar="SIZE",
            action=_StoreSizeTupleAction,
            help=(
                "a `width,height` for each image. A dimension can be -1 if "
                "no constraint should be applied"
            ),
        )
        parser.add_argument(
            "--min-size",
            metavar="MIN_SIZE",
            action=_StoreSizeTupleAction,
            help=(
                "a minimum `width,height` for each image. A dimension can be "
                "-1 if no constraint should be applied"
            ),
        )
        parser.add_argument(
            "--max-size",
            metavar="MAX_SIZE",
            action=_StoreSizeTupleAction,
            help=(
                "a maximum `width,height` for each image. A dimension can be "
                "-1 if no constraint should be applied"
            ),
        )
        parser.add_argument(
            "-i",
            "--interpolation",
            default=None,
            type=int,
            help="an optional `interpolation` argument for `cv2.resize()`",
        )
        parser.add_argument(
            "-e",
            "--ext",
            metavar="EXT",
            help="an image format to convert to (e.g., '.png' or '.jpg')",
        )
        parser.add_argument(
            "-f",
            "--force-reencode",
            action="store_true",
            help=(
                "whether to re-encode images whose parameters already meet "
                "the specified values"
            ),
        )
        parser.add_argument(
            "--media-field",
            metavar="MEDIA_FIELD",
            default="filepath",
            help="the input field containing the image paths to transform",
        )
        parser.add_argument(
            "--output-field",
            metavar="OUTPUT_FIELD",
            help=(
                "an optional field in which to store the paths to the "
                "transformed images. By default, `media_field` is updated "
                "in-place"
            ),
        )
        parser.add_argument(
            "--output-dir",
            metavar="OUTPUT_DIR",
            help=(
                "an optional output directory in which to write the "
                "transformed images. If none is provided, the images are "
                "updated in-place"
            ),
        )
        parser.add_argument(
            "--rel-dir",
            metavar="REL_DIR",
            help=(
                "an optional relative directory to strip from each input "
                "filepath to generate a unique identifier that is joined with "
                "`output_dir` to generate an output path for each image"
            ),
        )
        parser.add_argument(
            "-d",
            "--delete-originals",
            action="store_true",
            help="whether to delete the original images after transforming",
        )
        parser.add_argument(
            "-n",
            "--num-workers",
            default=None,
            type=int,
            help=(
                "the number of worker processes to use. The default is "
                "`multiprocessing.cpu_count()`"
            ),
        )
        parser.add_argument(
            "-s",
            "--skip-failures",
            action="store_true",
            help=(
                "whether to gracefully continue without raising an error if "
                "an image cannot be transformed"
            ),
        )

    @staticmethod
    def execute(parser, args):
        dataset = fod.load_dataset(args.name)
        foui.transform_images(
            dataset,
            size=args.size,
            min_size=args.min_size,
            max_size=args.max_size,
            interpolation=args.interpolation,
            ext=args.ext,
            force_reencode=args.force_reencode,
            media_field=args.media_field,
            output_field=args.output_field,
            output_dir=args.output_dir,
            rel_dir=args.rel_dir,
            delete_originals=args.delete_originals,
            num_workers=args.num_workers,
            skip_failures=args.skip_failures,
        )


class TransformVideosCommand(Command):
    """Transforms the videos in a dataset per the specified parameters.

    Examples::

        # Re-encode the videos in the dataset as H.264 MP4s
        fiftyone utils transform-videos <dataset-name> --reencode

        # Ensure that no videos in the dataset exceed 1920 x 1080 and 30fps
        fiftyone utils transform-videos <dataset-name> \\
            --max-size 1920,1080 --max-fps 30.0
    """

    @staticmethod
    def setup(parser):
        parser.add_argument(
            "name", metavar="DATASET_NAME", help="the name of the dataset"
        )
        parser.add_argument(
            "--fps",
            metavar="FPS",
            default=None,
            type=float,
            help="a frame rate at which to resample the videos",
        )
        parser.add_argument(
            "--min-fps",
            metavar="MIN_FPS",
            default=None,
            type=float,
            help=(
                "a minimum frame rate. Videos with frame rate below this "
                "value are upsampled"
            ),
        )
        parser.add_argument(
            "--max-fps",
            metavar="MAX_FPS",
            default=None,
            type=float,
            help=(
                "a maximum frame rate. Videos with frame rate exceeding this "
                "value are downsampled"
            ),
        )
        parser.add_argument(
            "--size",
            metavar="SIZE",
            action=_StoreSizeTupleAction,
            help=(
                "a `width,height` for each frame. A dimension can be -1 if "
                "no constraint should be applied"
            ),
        )
        parser.add_argument(
            "--min-size",
            metavar="MIN_SIZE",
            action=_StoreSizeTupleAction,
            help=(
                "a minimum `width,height` for each frame. A dimension can be "
                "-1 if no constraint should be applied"
            ),
        )
        parser.add_argument(
            "--max-size",
            metavar="MAX_SIZE",
            action=_StoreSizeTupleAction,
            help=(
                "a maximum `width,height` for each frame. A dimension can be "
                "-1 if no constraint should be applied"
            ),
        )
        parser.add_argument(
            "-r",
            "--reencode",
            action="store_true",
            help="whether to re-encode the videos as H.264 MP4s",
        )
        parser.add_argument(
            "-f",
            "--force-reencode",
            action="store_true",
            help=(
                "whether to re-encode videos whose parameters already meet "
                "the specified values"
            ),
        )
        parser.add_argument(
            "--media-field",
            metavar="MEDIA_FIELD",
            default="filepath",
            help="the input field containing the video paths to transform",
        )
        parser.add_argument(
            "--output-field",
            metavar="OUTPUT_FIELD",
            help=(
                "an optional field in which to store the paths to the "
                "transformed videos. By default, `media_field` is updated "
                "in-place"
            ),
        )
        parser.add_argument(
            "--output-dir",
            metavar="OUTPUT_DIR",
            help=(
                "an optional output directory in which to write the "
                "transformed videos. If none is provided, the videos are "
                "updated in-place"
            ),
        )
        parser.add_argument(
            "--rel-dir",
            metavar="REL_DIR",
            help=(
                "an optional relative directory to strip from each input "
                "filepath to generate a unique identifier that is joined with "
                "`output_dir` to generate an output path for each video"
            ),
        )
        parser.add_argument(
            "-d",
            "--delete-originals",
            action="store_true",
            help="whether to delete the original videos after transforming",
        )
        parser.add_argument(
            "-s",
            "--skip-failures",
            action="store_true",
            help=(
                "whether to gracefully continue without raising an error if "
                "a video cannot be transformed"
            ),
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="whether to log the `ffmpeg` commands that are executed",
        )

    @staticmethod
    def execute(parser, args):
        dataset = fod.load_dataset(args.name)
        fouv.transform_videos(
            dataset,
            fps=args.fps,
            min_fps=args.min_fps,
            max_fps=args.max_fps,
            size=args.size,
            min_size=args.min_size,
            max_size=args.max_size,
            reencode=args.reencode,
            force_reencode=args.force_reencode,
            media_field=args.media_field,
            output_field=args.output_field,
            output_dir=args.output_dir,
            rel_dir=args.rel_dir,
            delete_originals=args.delete_originals,
            skip_failures=args.skip_failures,
            verbose=args.verbose,
        )


def _print_dict_as_json(d):
    print(json.dumps(d, indent=4))


def _print_dict_as_table(d, headers=None):
    if headers is None:
        headers = ["key", "value"]

    records = [(k, v) for k, v in d.items()]
    table_str = tabulate(records, headers=headers, tablefmt=_TABLE_FORMAT)
    print(table_str)


def _has_subparsers(parser):
    for action in parser._actions:
        if isinstance(action, argparse._SubParsersAction):
            return True

    return False


def _iter_subparsers(parser):
    for action in parser._actions:
        if isinstance(action, argparse._SubParsersAction):
            for subparser in action.choices.values():
                yield subparser


class _RecursiveHelpAction(argparse._HelpAction):
    def __call__(self, parser, *args, **kwargs):
        self._recurse(parser)
        parser.exit()

    @staticmethod
    def _recurse(parser):
        print("\n%s\n%s" % ("*" * 79, parser.format_help()))
        for subparser in _iter_subparsers(parser):
            _RecursiveHelpAction._recurse(subparser)


class _ParseKwargsAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        kwargs = {}
        if not isinstance(values, list):
            values = [values]

        for value in values:
            if "=" not in value:
                key = value
                val = "True"
            else:
                key, val = value.split("=")

            kwargs[key.replace("-", "_")] = _parse_kwargs_value(val)

        setattr(namespace, self.dest, kwargs)


def _parse_kwargs_value(value):
    try:
        return int(value)
    except:
        pass

    try:
        return float(value)
    except:
        pass

    if value in ("True", "true"):
        return True

    if value in ("False", "false"):
        return False

    if value in ("None", ""):
        return None

    if "," in value:
        return [_parse_kwargs_value(v) for v in value.split(",")]

    return value


class _StoreSizeTupleAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        if value is not None:
            try:
                l, r = value.split(",")
                size = (int(l), int(r))
            except:
                raise ValueError(
                    "Invalid argument %s for parameter '%s'; expected "
                    "`width,height`" % (value, self.dest)
                )

        else:
            size = None

        setattr(namespace, self.dest, size)


def _register_main_command(command, version=None, recursive_help=True):
    parser = argparse.ArgumentParser(description=command.__doc__.rstrip())

    parser.set_defaults(execute=lambda args: command.execute(parser, args))
    command.setup(parser)

    if version:
        parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=version,
            help="show version info",
        )

    if recursive_help and _has_subparsers(parser):
        parser.add_argument(
            "--all-help",
            action=_RecursiveHelpAction,
            help="show help recursively and exit",
        )

    argcomplete.autocomplete(parser)
    return parser


def _register_command(parent, name, command, recursive_help=True):
    parser = parent.add_parser(
        name,
        help=command.__doc__.splitlines()[0],
        description=command.__doc__.rstrip(),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.set_defaults(execute=lambda args: command.execute(parser, args))
    command.setup(parser)

    if recursive_help and _has_subparsers(parser):
        parser.add_argument(
            "--all-help",
            action=_RecursiveHelpAction,
            help="show help recursively and exit",
        )

    return parser


def main():
    """Executes the `fiftyone` tool with the given command-line args."""
    parser = _register_main_command(FiftyOneCommand, version=foc.VERSION_LONG)
    args = parser.parse_args()
    args.execute(args)



"""
Context utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import json
import os
import typing as t

try:
    import IPython.display
except:
    pass


_COLAB = "COLAB"
_DATABRICKS = "DATABRICKS"
_IPYTHON = "IPYTHON"
_NONE = "NONE"

_DATABRICKS_PROXY = None
_DATABRICKS_HOST = None

_context = None


def is_notebook_context():
    """Determines whether this process is running in a notebook context, either
    Jupyter or Google Colab.

    Returns:
        True/False
    """
    return _get_context() != _NONE


def is_jupyter_context():
    """Determines whether this process is running in a Jupyter notebook.

    Returns:
        True/False
    """
    return _get_context() == _IPYTHON


def is_colab_context():
    """Determines whether this process is running in Google Colab.

    Returns:
        True/False
    """
    return _get_context() == _COLAB


def is_databricks_context():
    """Determines whether this process is running in Databricks.

    Returns:
        True/False
    """
    return _get_context() == _DATABRICKS


def _get_context():
    """Determine the most specific context that we're in.

    Returns:
        one of

        -   ``_COLAB``: we're in Colab with an IPython notebook context
        -   ``_DATABRICKS``: we're in Databricks with an IPython notebook context
        -   ``_IPYTHON``: we're in an IPython notebook context
            (e.g., from running `jupyter notebook` at the command line)
        -   ``_NONE``: we're in a non-notebook context, e.g., a Python script
            or a Python REPL
    """
    global _context
    if _context is not None:
        return _context

    if os.environ.get("FIFTYONE_CONTEXT", None):
        _context = os.environ["FIFTYONE_CONTEXT"]
        return _context

    # In Colab, the `google.colab` module is available, but the shell returned
    # by `IPython.get_ipython` does not have a `get_trait` method.
    try:
        import google.colab  # noqa: F401
        import IPython
    except ImportError:
        pass
    else:
        if IPython.get_ipython() is not None:
            # We'll assume that we're in a Colab notebook context.
            _context = _COLAB
            os.environ["FIFTYONE_CONTEXT"] = _context
            return _context

    # In Databricks, the `dbutils` module is available and required for the proxy,
    # but the shell returned by `IPython.get_ipython` does not return a kernel
    # via the `get_trait` method.
    try:
        # Location: /databricks/python_shell/dbruntime
        import IPython
        from dbruntime.dbutils import DBUtils  # noqa: F401
    except ImportError:
        pass
    else:
        if IPython.get_ipython() is not None:
            # We'll assume that we're in a Databricks notebook context.
            _context = _DATABRICKS
            os.environ["FIFTYONE_CONTEXT"] = _context
            return _context

    # In an IPython command line shell or Jupyter notebook, we can directly
    # query whether we're in a notebook context.
    try:
        import IPython
    except ImportError:
        pass
    else:
        ipython = IPython.get_ipython()
        if ipython is not None and ipython.has_trait("kernel"):
            _context = _IPYTHON
            os.environ["FIFTYONE_CONTEXT"] = _context
            return _context

    # Otherwise, we're not in a known notebook context.
    _context = _NONE
    os.environ["FIFTYONE_CONTEXT"] = _context

    return _context


def get_url(
    address: str,
    port: int,
    **kwargs: t.Dict[str, str],
) -> str:
    context = _get_context()
    if context == _COLAB:
        # pylint: disable=no-name-in-module,import-error
        from google.colab.output import eval_js

        _url = eval_js(f"google.colab.kernel.proxyPort({port})")
    elif _context == _DATABRICKS:
        _url = _get_databricks_proxy_url(port)
        kwargs["proxy"] = _get_databricks_proxy(port)
        kwargs["context"] = "databricks"
    else:
        _url = f"http://{address}:{port}/"

    params = "&".join([f"{k}={v}" for k, v in kwargs.items()])
    if params:
        _url = f"{_url}?{params}"

    return _url


def _get_databricks_proxy(port: int):
    _set_databricks()
    global _DATABRICKS_PROXY

    return f"{_DATABRICKS_PROXY}{port}/"


def _get_databricks_proxy_url(port: int):
    _set_databricks()
    global _DATABRICKS_HOST
    global _DATABRICKS_PROXY

    return f"https://{_DATABRICKS_HOST}{_DATABRICKS_PROXY}{port}/"


def _set_databricks() -> str:
    global _DATABRICKS_HOST
    global _DATABRICKS_PROXY

    if _DATABRICKS_PROXY:
        return

    shell = IPython.get_ipython()
    dbutils = shell.user_ns["dbutils"]
    data = json.loads(
        dbutils.entry_point.getDbutils().notebook().getContext().toJson()
    )["tags"]

    _DATABRICKS_HOST = data["browserHostName"]

    org_id = data["orgId"]
    cluster_id = data["clusterId"]
    _DATABRICKS_PROXY = f"/driver-proxy/o/{org_id}/{cluster_id}/"


class ContextError(EnvironmentError):
    """Exception raised when an action is taken in an unsupported context."""



"""
FiftyOne Zoo Datasets provided by ``tensorflow_datasets``.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import fiftyone.core.labels as fol
import fiftyone.core.utils as fou
import fiftyone.types as fot
import fiftyone.utils.imagenet as foui
import fiftyone.utils.data as foud
import fiftyone.zoo.datasets as fozd


_TFDS_IMPORT_ERROR = """

You tried to download a dataset from the FiftyOne Dataset Zoo using the
TensorFlow backend, but you do not have the necessary packages installed.

Ensure that you have `tensorflow` and `tensorflow_datasets` installed on your
machine, and then try running this command again.

See https://voxel51.com/docs/fiftyone/user_guide/dataset_zoo/index.html
for more information about working with the Dataset Zoo.
"""

_callback = lambda: fou.ensure_tfds(error_msg=_TFDS_IMPORT_ERROR)
tfds = fou.lazy_import("tensorflow_datasets", callback=_callback)


class TFDSDataset(fozd.ZooDataset):
    """Base class for zoo datasets that are provided via the
    ``tensorflow_datasets`` package.
    """

    pass


class MNISTDataset(TFDSDataset):
    """The MNIST database of handwritten digits.

    The dataset consists of 70,000 28 x 28 grayscale images in 10 classes.
    There are 60,000 training images and 10,000 test images.

    Example usage::

        import fiftyone as fo
        import fiftyone.zoo as foz

        dataset = foz.load_zoo_dataset("mnist", split="test")

        session = fo.launch_app(dataset)

    Dataset size
        21.00 MB

    Source
        http://yann.lecun.com/exdb/mnist
    """

    @property
    def name(self):
        return "mnist"

    @property
    def tags(self):
        return ("image", "classification")

    @property
    def supported_splits(self):
        return ("train", "test")

    def _download_and_prepare(self, dataset_dir, scratch_dir, split):
        def download_fcn(download_dir):
            return tfds.load(
                "mnist",
                split=split,
                data_dir=download_dir,
                download=True,
                with_info=True,
            )

        get_classes_fcn = lambda info: info.features["label"].names
        get_num_samples_fcn = lambda info: info.splits[split].num_examples
        sample_parser = _TFDSImageClassificationSampleParser()
        return _download_and_prepare(
            dataset_dir,
            scratch_dir,
            download_fcn,
            get_classes_fcn,
            get_num_samples_fcn,
            sample_parser,
        )


class FashionMNISTDataset(TFDSDataset):
    """The Fashion-MNIST database of Zalando's fashion article images.

    The dataset consists of 70,000 28 x 28 grayscale images in 10 classes.
    There are 60,000 training images and 10,000 test images.

    Example usage::

        import fiftyone as fo
        import fiftyone.zoo as foz

        dataset = foz.load_zoo_dataset("fashion-mnist", split="test")

        session = fo.launch_app(dataset)

    Dataset size
        36.42 MB

    Source
        https://github.com/zalandoresearch/fashion-mnist
    """

    @property
    def name(self):
        return "fashion-mnist"

    @property
    def tags(self):
        return ("image", "classification")

    @property
    def supported_splits(self):
        return ("train", "test")

    def _download_and_prepare(self, dataset_dir, scratch_dir, split):
        def download_fcn(download_dir):
            return tfds.load(
                "fashion_mnist",
                split=split,
                data_dir=download_dir,
                download=True,
                with_info=True,
            )

        get_classes_fcn = lambda info: info.features["label"].names
        get_num_samples_fcn = lambda info: info.splits[split].num_examples
        sample_parser = _TFDSImageClassificationSampleParser()
        return _download_and_prepare(
            dataset_dir,
            scratch_dir,
            download_fcn,
            get_classes_fcn,
            get_num_samples_fcn,
            sample_parser,
        )


class CIFAR10Dataset(TFDSDataset):
    """The CIFAR-10 dataset of images.

    The dataset consists of 60,000 32 x 32 color images in 10 classes, with
    6,000 images per class. There are 50,000 training images and 10,000 test
    images.

    Example usage::

        import fiftyone as fo
        import fiftyone.zoo as foz

        dataset = foz.load_zoo_dataset("cifar10", split="test")

        session = fo.launch_app(dataset)

    Dataset size
        132.40 MB

    Source
        https://www.cs.toronto.edu/~kriz/cifar.html
    """

    @property
    def name(self):
        return "cifar10"

    @property
    def tags(self):
        return ("image", "classification")

    @property
    def supported_splits(self):
        return ("train", "test")

    def _download_and_prepare(self, dataset_dir, scratch_dir, split):
        def download_fcn(download_dir):
            return tfds.load(
                "cifar10",
                split=split,
                data_dir=download_dir,
                download=True,
                with_info=True,
            )

        get_classes_fcn = lambda info: info.features["label"].names
        get_num_samples_fcn = lambda info: info.splits[split].num_examples
        sample_parser = _TFDSImageClassificationSampleParser()
        return _download_and_prepare(
            dataset_dir,
            scratch_dir,
            download_fcn,
            get_classes_fcn,
            get_num_samples_fcn,
            sample_parser,
        )


class CIFAR100Dataset(TFDSDataset):
    """The CIFAR-100 dataset of images.

    The dataset consists of 60,000 32 x 32 color images in 100 classes, with
    600 images per class. There are 50,000 training images and 10,000 test
    images.

    Example usage::

        import fiftyone as fo
        import fiftyone.zoo as foz

        dataset = foz.load_zoo_dataset("cifar100", split="test")

        session = fo.launch_app(dataset)

    Dataset size
        132.03 MB

    Source
        https://www.cs.toronto.edu/~kriz/cifar.html
    """

    @property
    def name(self):
        return "cifar100"

    @property
    def tags(self):
        return ("image", "classification")

    @property
    def supported_splits(self):
        return ("train", "test")

    def _download_and_prepare(self, dataset_dir, scratch_dir, split):
        def download_fcn(download_dir):
            return tfds.load(
                "cifar100",
                split=split,
                data_dir=download_dir,
                download=True,
                with_info=True,
            )

        get_classes_fcn = lambda info: info.features["label"].names
        get_num_samples_fcn = lambda info: info.splits[split].num_examples
        sample_parser = _TFDSImageClassificationSampleParser()
        return _download_and_prepare(
            dataset_dir,
            scratch_dir,
            download_fcn,
            get_classes_fcn,
            get_num_samples_fcn,
            sample_parser,
        )


class ImageNet2012Dataset(TFDSDataset):
    """The ImageNet 2012 dataset.

    ImageNet, as known as ILSVRC 2012, is an image dataset organized according
    to the WordNet hierarchy. Each meaningful concept in WordNet, possibly
    described by multiple words or word phrases, is called a "synonym set" or
    "synset". There are more than 100,000 synsets in WordNet, majority of them
    are nouns (80,000+). ImageNet provides on average 1,000 images to
    illustrate each synset. Images of each concept are quality-controlled and
    human-annotated. In its completion, we hope ImageNet will offer tens of
    millions of cleanly sorted images for most of the concepts in the WordNet
    hierarchy.

    Note that labels were never publicly released for the test set, so only the
    training and validation sets are provided.

    In order to load the ImageNet dataset, you must download the source data
    manually. The directory should be organized in the following format::

        source_dir/
            ILSVRC2012_devkit_t12.tar.gz    # both splits
            ILSVRC2012_img_train.tar        # train split
            ILSVRC2012_img_val.tar          # validation split

    You can register at http://www.image-net.org/download-images in order to
    get links to download the data.

    Example usage::

        import fiftyone as fo
        import fiftyone.zoo as foz

        # The path to the source files that you manually downloaded
        source_dir = "/path/to/dir-with-imagenet-files"

        dataset = foz.load_zoo_dataset(
            "imagenet-2012",
            split="validation",
            source_dir=source_dir,
        )

        session = fo.launch_app(dataset)

    Dataset size
        144.02 GB

    Source
        http://image-net.org

    Args:
        source_dir (None): the directory containing the manually downloaded
            ImageNet files
    """

    def __init__(self, source_dir=None):
        self.source_dir = source_dir

    @property
    def name(self):
        return "imagenet-2012"

    @property
    def tags(self):
        return ("image", "classification", "manual")

    @property
    def supported_splits(self):
        return ("train", "validation")

    @property
    def requires_manual_download(self):
        return True

    def _download_and_prepare(self, dataset_dir, scratch_dir, split):
        # Ensure that the source files have been manually downloaded
        foui.ensure_imagenet_manual_download(self.source_dir, split)

        if split == "validation":
            _split = "val"
        else:
            _split = split

        def download_fcn(_):
            return tfds.load(
                "imagenet2012",
                split=_split,
                data_dir=self.source_dir,
                with_info=True,
                download_and_prepare_kwargs={"download_dir": scratch_dir},
            )

        get_classes_fcn = lambda info: info.features["label"].names
        get_num_samples_fcn = lambda info: info.splits[_split].num_examples
        sample_parser = _TFDSImageClassificationSampleParser()
        return _download_and_prepare(
            dataset_dir,
            None,
            download_fcn,
            get_classes_fcn,
            get_num_samples_fcn,
            sample_parser,
        )


class VOC2007Dataset(TFDSDataset):
    """The dataset for the PASCAL Visual Object Classes Challenge 2007
    (VOC2007) for the detection competition.

    A total of 9,963 images are included in this dataset, where each image
    contains a set of objects, out of 20 different classes, making a total of
    24,640 annotated objects.

    Note that, as per the official dataset, the test set of VOC2007 does not
    contain annotations.

    Example usage::

        import fiftyone as fo
        import fiftyone.zoo as foz

        dataset = foz.load_zoo_dataset("voc-2007", split="validation")

        session = fo.launch_app(dataset)

    Dataset size
        868.85 MB

    Source
        http://host.robots.ox.ac.uk/pascal/VOC/voc2007
    """

    @property
    def name(self):
        return "voc-2007"

    @property
    def tags(self):
        return ("image", "detection")

    @property
    def supported_splits(self):
        return ("train", "validation", "test")

    def _download_and_prepare(self, dataset_dir, scratch_dir, split):
        def download_fcn(download_dir):
            return tfds.load(
                "voc/2007",
                split=split,
                data_dir=download_dir,
                download=True,
                with_info=True,
            )

        get_classes_fcn = lambda info: info.features["objects"]["label"].names
        get_num_samples_fcn = lambda info: info.splits[split].num_examples
        sample_parser = _TFDSImageDetectionSampleParser(
            bounding_box_field="bbox"
        )
        return _download_and_prepare(
            dataset_dir,
            scratch_dir,
            download_fcn,
            get_classes_fcn,
            get_num_samples_fcn,
            sample_parser,
        )


class VOC2012Dataset(TFDSDataset):
    """The dataset for the PASCAL Visual Object Classes Challenge 2012
    (VOC2012) for the detection competition.

    A total of 11,540 images are included in this dataset, where each image
    contains a set of objects, out of 20 different classes, making a total of
    27,450 annotated objects.

    Note that, as per the official dataset, the test set of VOC2012 does not
    contain annotations.

    Example usage::

        import fiftyone as fo
        import fiftyone.zoo as foz

        dataset = foz.load_zoo_dataset("voc-2012", split="validation")

        session = fo.launch_app(dataset)

    Dataset size
        3.59 GB

    Source
        http://host.robots.ox.ac.uk/pascal/VOC/voc2012
    """

    @property
    def name(self):
        return "voc-2012"

    @property
    def tags(self):
        return ("image", "detection")

    @property
    def supported_splits(self):
        return ("train", "validation", "test")

    def _download_and_prepare(self, dataset_dir, scratch_dir, split):
        def download_fcn(download_dir):
            return tfds.load(
                "voc/2012",
                split=split,
                data_dir=download_dir,
                download=True,
                with_info=True,
            )

        get_classes_fcn = lambda info: info.features["objects"]["label"].names
        get_num_samples_fcn = lambda info: info.splits[split].num_examples
        sample_parser = _TFDSImageDetectionSampleParser(
            bounding_box_field="bbox"
        )
        return _download_and_prepare(
            dataset_dir,
            scratch_dir,
            download_fcn,
            get_classes_fcn,
            get_num_samples_fcn,
            sample_parser,
        )


AVAILABLE_DATASETS = {
    "mnist": MNISTDataset,
    "fashion-mnist": FashionMNISTDataset,
    "cifar10": CIFAR10Dataset,
    "cifar100": CIFAR100Dataset,
    "imagenet-2012": ImageNet2012Dataset,
    "voc-2007": VOC2007Dataset,
    "voc-2012": VOC2012Dataset,
}


class _TFDSImageClassificationSampleParser(
    foud.ImageClassificationSampleParser
):
    def __init__(self, image_field="image", label_field="label", **kwargs):
        super().__init__(**kwargs)
        self.image_field = image_field
        self.label_field = label_field

    def _get_image(self):
        img = self.current_sample[self.image_field]
        return self._parse_image(img)

    def get_label(self):
        try:
            target = self.current_sample[self.label_field]
        except:
            return None

        return self._parse_label(target)


class _TFDSImageDetectionSampleParser(foud.ImageDetectionSampleParser):
    def __init__(self, image_field="image", objects_field="objects", **kwargs):
        super().__init__(**kwargs)
        self.image_field = image_field
        self.objects_field = objects_field

    def _get_image(self):
        img = self.current_sample[self.image_field]
        return self._parse_image(img)

    def get_label(self):
        try:
            target = self.current_sample[self.objects_field]
        except:
            return None

        if not self.normalized:
            # Absolute bounding box coordinates were provided, so we must have
            # the image to convert to relative coordinates
            img = self._current_image
        else:
            img = None

        return self._parse_label(target, img=img)

    def _parse_label(self, target, img=None):
        if target is None:
            return None

        # Convert from dict-of-lists to list-of-dicts
        target = [
            {self.bounding_box_field: bbox, self.label_field: label}
            for bbox, label in zip(
                target[self.bounding_box_field], target[self.label_field]
            )
        ]

        return super()._parse_label(target, img=img)

    def _parse_bbox(self, obj):
        # Format reference:
        # www.tensorflow.org/datasets/api_docs/python/tfds/features/BBoxFeature
        ymin, xmin, ymax, xmax = obj[self.bounding_box_field]
        w = xmax - xmin
        h = ymax - ymin
        return xmin, ymin, w, h


def _download_and_prepare(
    dataset_dir,
    scratch_dir,
    download_fcn,
    get_classes_fcn,
    get_num_samples_fcn,
    sample_parser,
):
    #
    # Download the TFDS dataset, if necessary
    #
    # Prevents ResourceExhaustedError that can arise...
    # https://github.com/tensorflow/datasets/issues/1441#issuecomment-581660890
    #
    with fou.ResourceLimit("RLIMIT_NOFILE", soft=4096, warn_on_failure=True):
        dataset, info = download_fcn(scratch_dir)

    classes = get_classes_fcn(info)
    num_samples = get_num_samples_fcn(info)
    sample_parser.classes = classes
    label_cls = sample_parser.label_cls

    if not isinstance(label_cls, (list, tuple)):
        label_cls = [label_cls]

    if fol.Classification in label_cls:
        dataset_type = fot.FiftyOneImageClassificationDataset()
        dataset_exporter = foud.FiftyOneImageClassificationDatasetExporter(
            dataset_dir, classes=classes
        )
    elif fol.Detections in label_cls:
        dataset_type = fot.FiftyOneImageDetectionDataset()
        dataset_exporter = foud.FiftyOneImageDetectionDatasetExporter(
            dataset_dir, classes=classes
        )
    else:
        raise ValueError("Unsupported sample parser %s" % type(sample_parser))

    try:
        samples = dataset.as_numpy_iterator()
    except AttributeError:
        # Must be tensorflow < 2.1
        samples = tfds.as_numpy(dataset)

    # Write the formatted dataset to `dataset_dir`
    foud.write_dataset(
        samples, sample_parser, dataset_exporter, num_samples=num_samples
    )

    return dataset_type, num_samples, classes



"""
Sphinx custom directives.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from docutils.parsers.rst import Directive, directives
from docutils.statemachine import StringList
from docutils import nodes


class CustomCardItemDirective(Directive):
    """A custom card item for use on pages that contain a list of links with
    short descriptions and optional images.

    The card item provides a heading, a description, and an optional image to
    the right. The entire card is clickable and links to the provided link.

    Example usage::

        .. customcarditem::
            :header: Custom card
            :description: This is a custom card. Click to learn more.
            :link: other/page.html
            :image: ../_static/images/custom-image.jpg
            :tags: Custom-Tag
    """

    option_spec = {
        "header": directives.unchanged,
        "description": directives.unchanged,
        "link": directives.unchanged,
        "image": directives.unchanged,
        "tags": directives.unchanged,
    }

    def run(self):
        header = self.options.get("header", "")
        description = self.options.get("description", "")
        link = self.options.get("link", "")
        image = self.options.get("image", "")
        tags = self.options.get("tags", "")

        if image:
            card_rst = _CUSTOM_CARD_TEMPLATE.format(
                header=header,
                description=description,
                link=link,
                image='<img src="%s">' % image,
                tags=tags,
            )
        else:
            # No image
            template = _CUSTOM_CARD_TEMPLATE.replace(
                '<div class="tutorials-image">{image}</div>', ""
            )

            card_rst = template.format(
                header=header,
                description=description,
                link=link,
                tags=tags,
            )

        card_list = StringList(card_rst.split("\n"))
        card = nodes.paragraph()
        self.state.nested_parse(card_list, self.content_offset, card)
        return [card]


_CUSTOM_CARD_TEMPLATE = """
.. raw:: html

    <div class="col-md-12 tutorials-card-container" data-tags={tags}>

    <div class="card tutorials-card" link={link}>

    <div class="card-body">

    <div class="card-title-container">
        <h4>{header}</h4>
    </div>

    <p class="card-summary">{description}</p>

    <p class="tags">{tags}</p>

    <div class="tutorials-image">{image}</div>

    </div>

    </div>

    </div>
"""


class CustomCalloutItemDirective(Directive):
    """A custom callout for use on table of contents-style pages that link into
    other pages.

    The callout contains a header, a body, a clickable button that links to
    the provided link, and an optional image.

    Example usage::

        .. customcalloutitem::
            :header: Custom header
            :description: Custom body
            :button_text: Custom button
            :button_link: other/page.html
            :image: ../_static/images/custom-image.jpg
    """

    option_spec = {
        "header": directives.unchanged,
        "description": directives.unchanged,
        "button_text": directives.unchanged,
        "button_link": directives.unchanged,
        "image": directives.unchanged,
    }

    def run(self):
        header = self.options.get("header", "")
        description = self.options.get("description", "")
        button_text = self.options.get("button_text", "")
        button_link = self.options.get("button_link", "")
        image = self.options.get("image", "")

        classes = "with-right-arrow" if button_link else ""
        attributes = (
            ""
            if button_link
            else 'onclick="return false;" style="pointer-events:none;cursor:default;"'
        )

        if image:
            callout_rst = _CUSTOM_CALLOUT_TEMPLATE.format(
                header=header,
                description=description,
                button_text=button_text,
                button_link=button_link,
                classes=classes,
                attributes=attributes,
                image='<img src="%s">' % image,
            )

        else:
            # No image
            template = _CUSTOM_CALLOUT_TEMPLATE.replace(
                "<div>{image}</div>", ""
            )

            callout_rst = template.format(
                header=header,
                description=description,
                button_text=button_text,
                button_link=button_link,
                classes=classes,
                attributes=attributes,
            )

        button_list = StringList(callout_rst.split("\n"))
        button = nodes.paragraph()
        self.state.nested_parse(button_list, self.content_offset, button)
        return [button]


_CUSTOM_CALLOUT_TEMPLATE = """
.. raw:: html

    <div class="col-md-6">
        <div class="text-container">
            <h3>{header}</h3>
            <p class="body-paragraph">{description}</p>
            <p><a class="btn {classes} callout-button" href="{button_link}"{attributes}>{button_text}</a></p>
            <div>{image}</div>
        </div>
    </div>
"""


class CustomButtonDirective(Directive):
    """A custom button for use on table of contents-style pages that link into
    other pages.

    The button is clickable and links to the provided link.

    Example usage::
        .. custombutton::
            :button_text: Custom button
            :button_link: other/page.html
    """

    option_spec = {
        "button_text": directives.unchanged,
        "button_link": directives.unchanged,
    }

    def run(self):
        button_text = self.options.get("button_text", "")
        button_link = self.options.get("button_link", "")

        classes = "with-right-arrow" if button_link else ""
        attributes = (
            ""
            if button_link
            else 'onclick="return false;" style="pointer-events:none;cursor:default;"'
        )

        callout_rst = _CUSTOM_BUTTON_TEMPLATE.format(
            button_text=button_text,
            button_link=button_link,
            classes=classes,
            attributes=attributes,
        )

        button_list = StringList(callout_rst.split("\n"))
        button = nodes.paragraph()
        self.state.nested_parse(button_list, self.content_offset, button)
        return [button]


_CUSTOM_BUTTON_TEMPLATE = """
.. raw:: html

    <div class="tutorials-callout-container">
        <a class="btn {classes} callout-button" href="{button_link}"{attributes}>{button_text}</a>
    </div>
"""


class CustomImageLinkDirective(Directive):
    """A custom image within a link nested in a div. Styling can be done via
    a parent container.


    Example usage::
        .. customimagelink::
            :image_link: other/page.html
            :image_src: images/image.png
            :image_src: My image
    """

    option_spec = {
        "image_link": directives.unchanged,
        "image_src": directives.unchanged,
        "image_title": directives.unchanged,
    }

    def run(self):
        image_link = self.options.get("image_link", "")
        image_src = self.options.get("image_src", "")
        image_title = self.options.get("image_title", "")

        callout_rst = _CUSTOM_IMAGE_LINK_TEMPLATE.format(
            image_link=image_link,
            image_src=image_src,
            image_title=image_title,
        )

        image_list = StringList(callout_rst.split("\n"))
        image = nodes.paragraph()
        self.state.nested_parse(image_list, self.content_offset, image)
        return [image]


_CUSTOM_IMAGE_LINK_TEMPLATE = """
.. raw:: html

    <div>
        <a href="{image_link}" title="{image_title}">
          <img src="{image_src}" alt="{image_title}"/>
        </a>
    </div>
"""



"""
Brain method runs framework.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from fiftyone.core.runs import Run, RunConfig, RunInfo, RunResults


class BrainInfo(RunInfo):
    """Information about an brain method that has been run on a dataset.

    Args:
        key: the brain key
        timestamp (None): the UTC ``datetime`` when the brain method was run
        config (None): the :class:`BrainMethodConfig` for the run
    """

    @classmethod
    def config_cls(cls):
        return BrainMethodConfig


class BrainMethodConfig(RunConfig):
    """Base class for configuring :class:`BrainMethod` instances.

    Args:
        **kwargs: any leftover keyword arguments after subclasses have done
            their parsing
    """

    pass


class BrainMethod(Run):
    """Base class for brain methods.

    Args:
        config: an :class:`BrainMethodConfig`
    """

    @classmethod
    def run_info_cls(cls):
        return BrainInfo

    @classmethod
    def _runs_field(cls):
        return "brain_methods"

    @classmethod
    def _run_str(cls):
        return "brain method run"

    @classmethod
    def _results_cache_field(cls):
        return "_brain_cache"


class BrainResults(RunResults):
    """Base class for brain method results."""

    pass



"""
Dataset format conversion utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import inspect
import logging
import os

import eta.core.utils as etau

import fiftyone.core.dataset as fod
import fiftyone.types as fot

from .exporters import build_dataset_exporter
from .importers import build_dataset_importer


logger = logging.getLogger(__name__)


def convert_dataset(
    input_dir=None,
    input_type=None,
    input_kwargs=None,
    dataset_importer=None,
    output_dir=None,
    output_type=None,
    output_kwargs=None,
    dataset_exporter=None,
    overwrite=False,
):
    """Converts a dataset stored on disk to another format on disk.

    The input dataset may be specified by providing either an ``input_dir`` and
    a corresponding ``input_type`` or by providing a ``dataset_importer``.

    The output dataset may be specified by providing either an ``output_dir``
    and a corresponding ``output_type`` or by providing a ``dataset_exporter``.

    Args:
        input_dir (None): the input dataset directory
        input_type (None): the :class:`fiftyone.types.Dataset` type of the
            dataset in ``input_dir``
        input_kwargs (None): optional kwargs dict to pass to the constructor of
            the :class:`fiftyone.utils.data.importers.DatasetImporter` for the
            ``input_type`` you specify
        dataset_importer (None): a
            :class:`fiftyone.utils.data.importers.DatasetImporter` to use to
            import the input dataset
        output_dir (None): the directory to which to write the output dataset
        output_type (None): the :class:`fiftyone.types.Dataset` type to write
            to ``output_dir``
        output_kwargs (None): optional kwargs dict to pass to the constructor
            of the :class:`fiftyone.utils.data.exporters.DatasetExporter` for
            the ``output_type`` you specify
        dataset_exporter (None): a
            :class:`fiftyone.utils.data.exporters.DatasetExporter` to use to
            export the dataset
        overwrite (False): whether to delete existing directories before
            performing the export (True) or to merge the export with existing
            files and directories (False)
    """
    if input_type is None and dataset_importer is None:
        raise ValueError(
            "Either `input_type` or `dataset_importer` must be provided"
        )

    if output_type is None and dataset_exporter is None:
        raise ValueError(
            "Either `output_type` or `dataset_exporter` must be provided"
        )

    dataset = fod.Dataset()
    images_dir = None

    try:
        # Build importer
        if dataset_importer is None:
            if input_kwargs is None:
                input_kwargs = {}

            # If the input dataset contains TFRecords, they must be unpacked
            # into a temporary directory during conversion
            if _is_tf_records(input_type) and "images_dir" not in input_kwargs:
                images_dir = etau.make_temp_dir()
                input_kwargs["images_dir"] = images_dir

            dataset_importer, _ = build_dataset_importer(
                input_type, dataset_dir=input_dir, **input_kwargs
            )

            logger.info(
                "Using input format: %s", etau.get_class_name(input_type)
            )
        else:
            input_dir = dataset_importer.input_dir
            logger.info(
                "Using importer: %s", etau.get_class_name(dataset_importer)
            )

        # Build exporter
        if dataset_exporter is None:
            if output_kwargs is None:
                output_kwargs = {}

            dataset_exporter, _ = build_dataset_exporter(
                output_type, export_dir=output_dir, **output_kwargs
            )
            logger.info(
                "Using export format: %s", etau.get_class_name(output_type)
            )
        else:
            output_dir = dataset_exporter.export_dir
            logger.info(
                "Using exporter: %s", etau.get_class_name(dataset_exporter)
            )

        # Import dataset
        if input_dir is not None:
            logger.info("Loading dataset from '%s'", input_dir)
        else:
            logger.info("Loading dataset")

        dataset.add_importer(dataset_importer)

        # Export dataset
        if output_dir is not None:
            logger.info("Exporting dataset to '%s'", output_dir)
        else:
            logger.info("Exporting dataset")

        dataset.export(dataset_exporter=dataset_exporter, overwrite=overwrite)
    finally:
        if images_dir is not None and os.path.isdir(images_dir):
            etau.delete_dir(images_dir)

        dataset.delete()


def _is_tf_records(dataset_type):
    if inspect.isclass(dataset_type):
        dataset_type = dataset_type()

    return isinstance(
        dataset_type,
        (fot.TFImageClassificationDataset, fot.TFObjectDetectionDataset),
    )



"""
Database utilities.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import atexit
from datetime import datetime
import logging
from multiprocessing.pool import ThreadPool
import os

import asyncio
from bson import json_util
from bson.codec_options import CodecOptions
from mongoengine import connect
import mongoengine.errors as moe
import motor.motor_asyncio as mtr

from packaging.version import Version
import pymongo
from pymongo.errors import BulkWriteError, ServerSelectionTimeoutError
import pytz

import eta.core.utils as etau

import fiftyone as fo
import fiftyone.constants as foc
from fiftyone.core.config import FiftyOneConfigError
import fiftyone.core.fields as fof
import fiftyone.core.service as fos
import fiftyone.core.utils as fou

from .document import Document

fod = fou.lazy_import("fiftyone.core.dataset")

logger = logging.getLogger(__name__)

_client = None
_async_client = None
_connection_kwargs = {}
_db_service = None


#
# IMPORTANT DATABASE CONFIG REQUIREMENTS
#
# All past and future versions of FiftyOne must be able to deduce the
# database's current version and type from the `config` collection without an
# error being raised so that migrations can be properly run and, if necsssary,
# informative errors can be raised alerting the user that they are using the
# wrong version or type of client.
#
# This is currently guaranteed because:
#   - `DatabaseConfigDocument` is declared as non-strict, so any past or future
#     fields that are not currently defined will not cause an error
#   - All declared fields are optional and we have promised ourselves that
#     their type and meaning will never change
#


class DatabaseConfigDocument(Document):
    """Backing document for the database config."""

    # strict=False lets this class ignore unknown fields from other versions
    meta = {"collection": "config", "strict": False}

    version = fof.StringField()
    type = fof.StringField()


def get_db_config():
    """Retrieves the database config.

    Returns:
        a :class:`DatabaseConfigDocument`
    """
    save = False

    try:
        # pylint: disable=no-member
        config = DatabaseConfigDocument.objects.get()
    except moe.DoesNotExist:
        config = DatabaseConfigDocument()
        save = True

    if config.version is None:
        #
        # The database version was added to this config in v0.15.0, so if no
        # version is available, assume the database is at the preceeding
        # release. It's okay if the database's version is actually older,
        # because there are no significant admin migrations prior to v0.14.4.
        #
        # This needs to be implemented here rather than in a migration because
        # this information is required in order to run migrations...
        #
        config.version = "0.14.4"
        save = True

    if config.type is None:
        #
        # If the database has no type, then assume the type of the client that
        # is currently connecting to it.
        #
        # This needs to be implemented here rather than in a migration because
        # this information is required in order to decide if a client is
        # allowed to connect to the database at all (a precursor to running a
        # migration)...
        #
        config.type = foc.CLIENT_TYPE
        save = True

    if save:
        config.save()

    return config


def establish_db_conn(config):
    """Establishes the database connection.

    If ``fiftyone.config.database_uri`` is defined, then we connect to that
    URI. Otherwise, a :class:`fiftyone.core.service.DatabaseService` is
    created.

    Args:
        config: a :class:`fiftyone.core.config.FiftyOneConfig`

    Raises:
        ConnectionError: if a connection to ``mongod`` could not be established
        FiftyOneConfigError: if ``fiftyone.config.database_uri`` is not
            defined and ``mongod`` could not be found
        ServiceExecutableNotFound: if
            :class:`fiftyone.core.service.DatabaseService` startup was
            attempted, but ``mongod`` was not found in :mod:`fiftyone.db.bin`
        RuntimeError: if the ``mongod`` found does not meet FiftyOne's
            requirements, or validation could not occur
    """
    global _client
    global _db_service
    global _connection_kwargs

    established_port = os.environ.get("FIFTYONE_PRIVATE_DATABASE_PORT", None)
    if established_port is not None:
        _connection_kwargs["port"] = int(established_port)
    if config.database_uri is not None:
        _connection_kwargs["host"] = config.database_uri
    elif _db_service is None:
        if os.environ.get("FIFTYONE_DISABLE_SERVICES", False):
            return

        try:
            _db_service = fos.DatabaseService()
            port = _db_service.port
            _connection_kwargs["port"] = port
            os.environ["FIFTYONE_PRIVATE_DATABASE_PORT"] = str(port)

        except fos.ServiceExecutableNotFound as error:
            if fou.is_32_bit():
                raise FiftyOneConfigError(
                    "MongoDB is not supported on 32-bit systems. Please "
                    "define a `database_uri` in your "
                    "`fiftyone.core.config.FiftyOneConfig` to define a "
                    "connection to your own MongoDB instance or cluster "
                )

            raise error

    _client = pymongo.MongoClient(
        **_connection_kwargs, appname=foc.DATABASE_APPNAME
    )
    _validate_db_version(config, _client)

    # Register cleanup method
    atexit.register(_delete_non_persistent_datasets_if_allowed)

    connect(config.database_name, **_connection_kwargs)

    config = get_db_config()
    if foc.CLIENT_TYPE != config.type:
        raise ConnectionError(
            "Cannot connect to database type '%s' with client type '%s'"
            % (config.type, foc.CLIENT_TYPE)
        )


def _connect():
    global _client
    if _client is None:
        global _connection_kwargs

        _client = pymongo.MongoClient(
            **_connection_kwargs, appname=foc.DATABASE_APPNAME
        )
        connect(fo.config.database_name, **_connection_kwargs)


def _async_connect():
    global _async_client
    if _async_client is None:
        global _connection_kwargs
        _async_client = mtr.AsyncIOMotorClient(
            **_connection_kwargs, appname=foc.DATABASE_APPNAME
        )


def _delete_non_persistent_datasets_if_allowed():
    """Deletes all non-persistent datasets if and only if we are the only
    client currently connected to the database.
    """
    try:
        num_connections = len(
            list(
                _client.admin.aggregate(
                    [
                        {"$currentOp": {"allUsers": True}},
                        {"$project": {"appName": 1, "command": 1}},
                        {
                            "$match": {
                                "appName": foc.DATABASE_APPNAME,
                                "command.ismaster": 1,
                            }
                        },
                    ]
                )
            )
        )
    except:
        logger.warning(
            "Skipping automatic non-persistent dataset cleanup. This action "
            "requires read access of the 'admin' database"
        )
        return

    try:
        if num_connections <= 1:
            fod.delete_non_persistent_datasets()
    except:
        logger.exception("Skipping automatic non-persistent dataset cleanup")


def _validate_db_version(config, client):
    try:
        version = Version(client.server_info()["version"])
    except Exception as e:
        if isinstance(e, ServerSelectionTimeoutError):
            raise ConnectionError("Could not connect to `mongod`") from e

        raise RuntimeError("Failed to validate `mongod` version") from e

    if config.database_validation and version < foc.MIN_MONGODB_VERSION:
        raise RuntimeError(
            "Found `mongod` version %s, but only %s and higher are "
            "compatible. You can suppress this exception by setting your "
            "`database_validation` config parameter to `False`. See "
            "https://voxel51.com/docs/fiftyone/user_guide/config.html#configuring-a-mongodb-connection "
            "for more information" % (version, foc.MIN_MONGODB_VERSION)
        )


def aggregate(collection, pipelines):
    """Executes one or more aggregations on a collection.

    Multiple aggregations are executed using multiple threads, and their
    results are returned as lists rather than cursors.

    Args:
        collection: a ``pymongo.collection.Collection`` or
            ``motor.motor_asyncio.AsyncIOMotorCollection``
        pipelines: a MongoDB aggregation pipeline or a list of pipelines

    Returns:
        -   If a single pipeline is provided, a
            ``pymongo.command_cursor.CommandCursor`` or
            ``motor.motor_asyncio.AsyncIOMotorCommandCursor`` is returned

        -   If multiple pipelines are provided, each cursor is extracted into
            a list and the list of lists is returned
    """
    pipelines = list(pipelines)

    is_list = pipelines and not isinstance(pipelines[0], dict)
    if not is_list:
        pipelines = [pipelines]

    num_pipelines = len(pipelines)
    if isinstance(collection, mtr.AsyncIOMotorCollection):
        if num_pipelines == 1 and not is_list:
            return collection.aggregate(pipelines[0], allowDiskUse=True)

        return _do_async_pooled_aggregate(collection, pipelines)

    if num_pipelines == 1:
        result = collection.aggregate(pipelines[0], allowDiskUse=True)
        return [result] if is_list else result

    return _do_pooled_aggregate(collection, pipelines)


def _do_pooled_aggregate(collection, pipelines):
    # @todo: MongoDB 5.0 supports snapshots which can be used to make the
    # results consistent, i.e. read from the same point in time
    with ThreadPool(processes=len(pipelines)) as pool:
        return pool.map(
            lambda p: list(collection.aggregate(p, allowDiskUse=True)),
            pipelines,
            chunksize=1,
        )


async def _do_async_pooled_aggregate(collection, pipelines):
    return await asyncio.gather(
        *[_do_async_aggregate(collection, pipeline) for pipeline in pipelines]
    )


async def _do_async_aggregate(collection, pipeline):
    return [i async for i in collection.aggregate(pipeline, allowDiskUse=True)]


def get_db_client():
    """Returns a database client.

    Returns:
        a ``pymongo.mongo_client.MongoClient``
    """
    _connect()
    return _client


def get_db_conn():
    """Returns a connection to the database.

    Returns:
        a ``pymongo.database.Database``
    """
    _connect()
    db = _client[fo.config.database_name]
    return _apply_options(db)


def get_async_db_client():
    """Returns an async database client.

    Returns:
        a ``motor.motor_asyncio.AsyncIOMotorClient``
    """
    _async_connect()
    return _async_client


def get_async_db_conn():
    """Returns an async connection to the database.

    Returns:
        a ``motor.motor_asyncio.AsyncIOMotorDatabase``
    """
    db = get_async_db_client()[fo.config.database_name]
    return _apply_options(db)


def _apply_options(db):
    timezone = fo.config.timezone

    if not timezone or timezone.lower() == "utc":
        return db

    if timezone.lower() == "local":
        tzinfo = datetime.now().astimezone().tzinfo
    else:
        tzinfo = pytz.timezone(timezone)

    return db.with_options(
        codec_options=CodecOptions(tz_aware=True, tzinfo=tzinfo)
    )


def drop_database():
    """Drops the database."""
    _connect()
    _client.drop_database(fo.config.database_name)


def sync_database():
    """Syncs all pending database writes to disk."""
    if _client is not None:
        _client.admin.command("fsync")


def list_collections():
    """Returns a list of all collection names in the database.

    Returns:
        a list of all collection names
    """
    conn = get_db_conn()
    return list(conn.list_collection_names())


def drop_collection(collection_name):
    """Drops specified collection from the database.

    Args:
        collection_name: the collection name
    """
    conn = get_db_conn()
    conn.drop_collection(collection_name)


def drop_orphan_collections(dry_run=False):
    """Drops all orphan collections from the database.

    Orphan collections are collections that are not associated with any known
    dataset or other collections used by FiftyOne.

    Args:
        dry_run (False): whether to log the actions that would be taken but not
            perform them
    """
    conn = get_db_conn()
    _logger = _get_logger(dry_run=dry_run)

    colls_in_use = set()
    for dataset_dict in conn.datasets.find({}):
        sample_coll_name = dataset_dict.get("sample_collection_name", None)
        if sample_coll_name:
            colls_in_use.add(sample_coll_name)
            colls_in_use.add("frames." + sample_coll_name)

    # Only collections with these prefixes may be deleted
    coll_prefixes = ("samples.", "frames.", "patches.", "clips.")

    for coll_name in conn.list_collection_names():
        if coll_name not in colls_in_use and any(
            coll_name.startswith(prefix) for prefix in coll_prefixes
        ):
            _logger.info("Dropping collection '%s'", coll_name)
            if not dry_run:
                conn.drop_collection(coll_name)


def drop_orphan_run_results(dry_run=False):
    """Drops all orphan run results from the database.

    Orphan run results are results that are not associated with any known
    dataset.

    Args:
        dry_run (False): whether to log the actions that would be taken but not
            perform them
    """
    conn = get_db_conn()
    _logger = _get_logger(dry_run=dry_run)

    results_in_use = set()
    for name in list_datasets():
        dataset_dict = conn.datasets.find_one({"name": name})
        results_in_use.update(_get_result_ids(dataset_dict))

    all_run_results = set(conn.fs.files.distinct("_id", {}, {}))

    orphan_results = [
        _id for _id in all_run_results if _id not in results_in_use
    ]

    if not orphan_results:
        return

    _logger.info(
        "Deleting %d orphan run result(s): %s",
        len(orphan_results),
        orphan_results,
    )
    if not dry_run:
        _delete_run_results(orphan_results)


def stream_collection(collection_name):
    """Streams the contents of the collection to stdout.

    Args:
        collection_name: the name of the collection
    """
    conn = get_db_conn()
    coll = conn[collection_name]
    objects = map(fou.pformat, coll.find({}))
    fou.stream_objects(objects)


def get_collection_stats(collection_name):
    """Sets stats about the collection.

    Args:
        collection_name: the name of the collection

    Returns:
        a stats dict
    """
    conn = get_db_conn()
    stats = dict(conn.command("collstats", collection_name))
    stats["wiredTiger"] = None
    stats["indexDetails"] = None
    return stats


def count_documents(coll, pipeline):
    result = aggregate(coll, pipeline + [{"$count": "count"}])

    try:
        return list(result)[0]["count"]
    except:
        pass

    return 0


def export_document(doc, json_path):
    """Exports the document to disk in JSON format.

    Args:
        doc: a BSON document dict
        json_path: the path to write the JSON file
    """
    etau.write_file(json_util.dumps(doc), json_path)


def export_collection(
    docs,
    json_dir_or_path,
    key="documents",
    patt="{idx:06d}-{id}.json",
    num_docs=None,
):
    """Exports the collection to disk in JSON format.

    Args:
        docs: an iterable containing the documents to export
        json_dir_or_path: the path to write a single JSON file containing the
            entire collection, or a directory in which to write per-document
            JSON files
        key ("documents"): the field name under which to store the documents
            when ``json_path`` is a single JSON file
        patt ("{idx:06d}-{id}.json"): a filename pattern to use when
            ``json_path`` is a directory. The pattern may contain ``idx`` to
            refer to the index of the document in ``docs`` or ``id`` to refer
            to the document's ID
        num_docs (None): the total number of documents. If omitted, this must
            be computable via ``len(docs)``
    """
    if num_docs is None:
        num_docs = len(docs)

    if json_dir_or_path.endswith(".json"):
        _export_collection_single(docs, json_dir_or_path, key, num_docs)
    else:
        _export_collection_multi(docs, json_dir_or_path, patt, num_docs)


def _export_collection_single(docs, json_path, key, num_docs):
    etau.ensure_basedir(json_path)

    with open(json_path, "w") as f:
        f.write('{"%s": [' % key)
        with fou.ProgressBar(total=num_docs, iters_str="docs") as pb:
            for idx, doc in pb(enumerate(docs, 1)):
                f.write(json_util.dumps(doc))
                if idx < num_docs:
                    f.write(",")

        f.write("]}")


def _export_collection_multi(docs, json_dir, patt, num_docs):
    etau.ensure_dir(json_dir)

    json_patt = os.path.join(json_dir, patt)
    with fou.ProgressBar(total=num_docs, iters_str="docs") as pb:
        for idx, doc in pb(enumerate(docs, 1)):
            json_path = json_patt.format(idx=idx, id=str(doc["_id"]))
            export_document(doc, json_path)


def import_document(json_path):
    """Imports a document from JSON on disk.

    Args:
        json_path: the path to the document

    Returns:
        a BSON document dict
    """
    with open(json_path, "r") as f:
        return json_util.loads(f.read())


def import_collection(json_dir_or_path, key="documents"):
    """Imports the collection from JSON on disk.

    Args:
        json_dir_or_path: the path to a JSON file on disk, or a directory
            containing per-document JSON files
        key ("documents"): the field name under which the documents are stored
            when ``json_path`` is a single JSON file

    Returns:
        a tuple of

        -   an iterable of BSON documents
        -   the number of documents
    """
    if json_dir_or_path.endswith(".json"):
        return _import_collection_single(json_dir_or_path, key)

    return _import_collection_multi(json_dir_or_path)


def _import_collection_single(json_path, key):
    with open(json_path, "r") as f:
        docs = json_util.loads(f.read()).get(key, [])

    num_docs = len(docs)

    return docs, num_docs


def _import_collection_multi(json_dir):
    json_paths = [
        p
        for p in etau.list_files(json_dir, abs_paths=True)
        if p.endswith(".json")
    ]
    docs = map(import_document, json_paths)

    return docs, len(json_paths)


def insert_documents(docs, coll, ordered=False, progress=False, num_docs=None):
    """Inserts documents into a collection.

    The ``_id`` field of the input documents will be populated if it is not
    already set.

    Args:
        docs: an iterable of BSON document dicts
        coll: a pymongo collection
        ordered (False): whether the documents must be inserted in order
        progress (False): whether to render a progress bar tracking the
            insertion
        num_docs (None): the total number of documents. Only used when
            ``progress=True``. If omitted, this will be computed via
            ``len(docs)``, if possible

    Returns:
        a list of IDs of the inserted documents
    """
    ids = []

    try:
        batcher = fou.DynamicBatcher(
            docs,
            target_latency=0.2,
            init_batch_size=1,
            max_batch_beta=2.0,
            max_batch_size=100000,  # mongodb limit
            progress=progress,
            total=num_docs,
        )

        with batcher:
            for batch in batcher:
                batch = list(batch)
                coll.insert_many(batch, ordered=ordered)
                ids.extend(b["_id"] for b in batch)
    except BulkWriteError as bwe:
        msg = bwe.details["writeErrors"][0]["errmsg"]
        raise ValueError(msg) from bwe

    return ids


def bulk_write(ops, coll, ordered=False):
    """Performs a batch of write operations on a collection.

    Args:
        ops: a list of pymongo operations
        coll: a pymongo collection
        ordered (False): whether the operations must be performed in order
    """
    try:
        for ops_batch in fou.iter_batches(ops, 100000):  # mongodb limit
            coll.bulk_write(list(ops_batch), ordered=ordered)
    except BulkWriteError as bwe:
        msg = bwe.details["writeErrors"][0]["errmsg"]
        raise ValueError(msg) from bwe


def list_datasets():
    """Returns the list of available FiftyOne datasets.

    This is a low-level implementation of dataset listing that does not call
    :meth:`fiftyone.core.dataset.list_datasets`, which is helpful if a
    database may be corrupted.

    Returns:
        a list of :class:`Dataset` names
    """
    conn = get_db_conn()
    return conn.datasets.distinct("name")


def delete_dataset(name, dry_run=False):
    """Deletes the dataset with the given name.

    This is a low-level implementation of deletion that does not call
    :meth:`fiftyone.core.dataset.load_dataset`, which is helpful if a dataset's
    backing document or collections are corrupted and cannot be loaded via the
    normal pathways.

    Args:
        name: the name of the dataset
        dry_run (False): whether to log the actions that would be taken but not
            perform them
    """
    conn = get_db_conn()
    _logger = _get_logger(dry_run=dry_run)

    dataset_dict = conn.datasets.find_one({"name": name})
    if not dataset_dict:
        _logger.warning("Dataset '%s' not found", name)
        return

    _logger.info("Dropping document '%s' from 'datasets' collection", name)
    if not dry_run:
        conn.datasets.delete_one({"name": name})

    if "sample_collection_name" not in dataset_dict:
        _logger.warning(
            "Cannot find sample/frame collections for dataset '%s'; stopping "
            "now. Use `drop_orphan_collections()` to cleanup any dangling "
            "collections",
            name,
        )
        return

    collections = conn.list_collection_names()

    sample_collection_name = dataset_dict["sample_collection_name"]
    if sample_collection_name in collections:
        _logger.info("Dropping collection '%s'", sample_collection_name)
        if not dry_run:
            conn.drop_collection(sample_collection_name)

    frame_collection_name = "frames." + sample_collection_name
    if frame_collection_name in collections:
        _logger.info("Dropping collection '%s'", frame_collection_name)
        if not dry_run:
            conn.drop_collection(frame_collection_name)

    delete_results = _get_result_ids(dataset_dict)

    if delete_results:
        _logger.info("Deleting %d run result(s)", len(delete_results))
        if not dry_run:
            _delete_run_results(delete_results)


def delete_annotation_run(name, anno_key, dry_run=False):
    """Deletes the annotation run with the given key from the dataset with
    the given name.

    This is a low-level implementation of deletion that does not call
    :meth:`fiftyone.core.dataset.load_dataset` or
    :meth:`fiftyone.core.collections.SampleCollection.delete_annotation_run`,
    which is helpful if a dataset's backing document or collections are
    corrupted and cannot be loaded via the normal pathways.

    Note that, as this method does not load :class:`fiftyone.core.runs.Run`
    instances, it does not call :meth:`fiftyone.core.runs.Run.cleanup`.

    Args:
        name: the name of the dataset
        anno_key: the annotation key
        dry_run (False): whether to log the actions that would be taken but not
            perform them
    """
    conn = get_db_conn()
    _logger = _get_logger(dry_run=dry_run)

    dataset_dict = conn.datasets.find_one({"name": name})
    if not dataset_dict:
        _logger.warning("Dataset '%s' not found", name)
        return

    annotation_runs = dataset_dict.get("annotation_runs", {})
    if anno_key not in annotation_runs:
        _logger.warning(
            "Dataset '%s' has no annotation run with key '%s'",
            name,
            anno_key,
        )
        return

    run_doc = annotation_runs.pop(anno_key)
    result_id = run_doc.get("results", None)

    if result_id is not None:
        _logger.info("Deleting run result '%s'", result_id)
        if not dry_run:
            _delete_run_results([result_id])

    _logger.info(
        "Deleting annotation run '%s' from dataset '%s'", anno_key, name
    )
    if not dry_run:
        conn.datasets.replace_one({"name": name}, dataset_dict)


def delete_annotation_runs(name, dry_run=False):
    """Deletes all annotation runs from the dataset with the given name.

    This is a low-level implementation of deletion that does not call
    :meth:`fiftyone.core.dataset.load_dataset` or
    :meth:`fiftyone.core.collections.SampleCollection.delete_annotation_runs`,
    which is helpful if a dataset's backing document or collections are
    corrupted and cannot be loaded via the normal pathways.

    Note that, as this method does not load :class:`fiftyone.core.runs.Run`
    instances, it does not call :meth:`fiftyone.core.runs.Run.cleanup`.

    Args:
        name: the name of the dataset
        dry_run (False): whether to log the actions that would be taken but not
            perform them
    """
    conn = get_db_conn()
    _logger = _get_logger(dry_run=dry_run)

    dataset_dict = conn.datasets.find_one({"name": name})
    if not dataset_dict:
        _logger.warning("Dataset '%s' not found", name)
        return

    anno_keys = []
    result_ids = []
    for anno_key, run_doc in dataset_dict.get("annotation_runs", {}).items():
        anno_keys.append(anno_key)

        result_id = run_doc.get("results", None)
        if result_id is not None:
            result_ids.append(result_id)

    if result_ids:
        _logger.info("Deleting %d run result(s)", len(result_ids))
        if not dry_run:
            _delete_run_results(result_ids)

    _logger.info(
        "Deleting annotation runs %s from dataset '%s'", anno_keys, name
    )
    if not dry_run:
        dataset_dict["annotation_runs"] = {}
        conn.datasets.replace_one({"name": name}, dataset_dict)


def delete_brain_run(name, brain_key, dry_run=False):
    """Deletes the brain method run with the given key from the dataset with
    the given name.

    This is a low-level implementation of deletion that does not call
    :meth:`fiftyone.core.dataset.load_dataset` or
    :meth:`fiftyone.core.collections.SampleCollection.delete_brain_run`,
    which is helpful if a dataset's backing document or collections are
    corrupted and cannot be loaded via the normal pathways.

    Note that, as this method does not load :class:`fiftyone.core.runs.Run`
    instances, it does not call :meth:`fiftyone.core.runs.Run.cleanup`.

    Args:
        name: the name of the dataset
        brain_key: the brain key
        dry_run (False): whether to log the actions that would be taken but not
            perform them
    """
    conn = get_db_conn()
    _logger = _get_logger(dry_run=dry_run)

    dataset_dict = conn.datasets.find_one({"name": name})
    if not dataset_dict:
        _logger.warning("Dataset '%s' not found", name)
        return

    brain_methods = dataset_dict.get("brain_methods", {})
    if brain_key not in brain_methods:
        _logger.warning(
            "Dataset '%s' has no brain method run with key '%s'",
            name,
            brain_key,
        )
        return

    run_doc = brain_methods.pop(brain_key)
    result_id = run_doc.get("results", None)

    if result_id is not None:
        _logger.info("Deleting run result '%s'", result_id)
        if not dry_run:
            _delete_run_results([result_id])

    _logger.info(
        "Deleting brain method run '%s' from dataset '%s'", brain_key, name
    )
    if not dry_run:
        conn.datasets.replace_one({"name": name}, dataset_dict)


def delete_brain_runs(name, dry_run=False):
    """Deletes all brain method runs from the dataset with the given name.

    This is a low-level implementation of deletion that does not call
    :meth:`fiftyone.core.dataset.load_dataset` or
    :meth:`fiftyone.core.collections.SampleCollection.delete_brain_runs`,
    which is helpful if a dataset's backing document or collections are
    corrupted and cannot be loaded via the normal pathways.

    Note that, as this method does not load :class:`fiftyone.core.runs.Run`
    instances, it does not call :meth:`fiftyone.core.runs.Run.cleanup`.

    Args:
        name: the name of the dataset
        dry_run (False): whether to log the actions that would be taken but not
            perform them
    """
    conn = get_db_conn()
    _logger = _get_logger(dry_run=dry_run)

    dataset_dict = conn.datasets.find_one({"name": name})
    if not dataset_dict:
        _logger.warning("Dataset '%s' not found", name)
        return

    brain_keys = []
    result_ids = []
    for brain_key, run_doc in dataset_dict.get("brain_methods", {}).items():
        brain_keys.append(brain_key)

        result_id = run_doc.get("results", None)
        if result_id is not None:
            result_ids.append(result_id)

    if result_ids:
        _logger.info("Deleting %d run result(s)", len(result_ids))
        if not dry_run:
            _delete_run_results(result_ids)

    _logger.info(
        "Deleting brain method runs %s from dataset '%s'",
        brain_keys,
        name,
    )
    if not dry_run:
        dataset_dict["brain_methods"] = {}
        conn.datasets.replace_one({"name": name}, dataset_dict)


def delete_evaluation(name, eval_key, dry_run=False):
    """Deletes the evaluation run with the given key from the dataset with the
    given name.

    This is a low-level implementation of deletion that does not call
    :meth:`fiftyone.core.dataset.load_dataset` or
    :meth:`fiftyone.core.collections.SampleCollection.delete_evaluation`,
    which is helpful if a dataset's backing document or collections are
    corrupted and cannot be loaded via the normal pathways.

    Note that, as this method does not load :class:`fiftyone.core.runs.Run`
    instances, it does not call :meth:`fiftyone.core.runs.Run.cleanup`.

    Args:
        name: the name of the dataset
        eval_key: the evaluation key
        dry_run (False): whether to log the actions that would be taken but not
            perform them
    """
    conn = get_db_conn()
    _logger = _get_logger(dry_run=dry_run)

    dataset_dict = conn.datasets.find_one({"name": name})
    if not dataset_dict:
        _logger.warning("Dataset '%s' not found", name)
        return

    evaluations = dataset_dict.get("evaluations", {})
    if eval_key not in evaluations:
        _logger.warning(
            "Dataset '%s' has no evaluation with key '%s'", name, eval_key
        )
        return

    run_doc = evaluations.pop(eval_key)
    result_id = run_doc.get("results", None)

    if result_id is not None:
        _logger.info("Deleting run result '%s'", result_id)
        if not dry_run:
            _delete_run_results([result_id])

    _logger.info("Deleting evaluation '%s' from dataset '%s'", eval_key, name)
    if not dry_run:
        conn.datasets.replace_one({"name": name}, dataset_dict)


def delete_evaluations(name, dry_run=False):
    """Deletes all evaluations from the dataset with the given name.

    This is a low-level implementation of deletion that does not call
    :meth:`fiftyone.core.dataset.load_dataset` or
    :meth:`fiftyone.core.collections.SampleCollection.delete_evaluations`,
    which is helpful if a dataset's backing document or collections are
    corrupted and cannot be loaded via the normal pathways.

    Note that, as this method does not load :class:`fiftyone.core.runs.Run`
    instances, it does not call :meth:`fiftyone.core.runs.Run.cleanup`.

    Args:
        name: the name of the dataset
        dry_run (False): whether to log the actions that would be taken but not
            perform them
    """
    conn = get_db_conn()
    _logger = _get_logger(dry_run=dry_run)

    dataset_dict = conn.datasets.find_one({"name": name})
    if not dataset_dict:
        _logger.warning("Dataset '%s' not found", name)
        return

    eval_keys = []
    result_ids = []
    for eval_key, run_doc in dataset_dict.get("evaluations", {}).items():
        eval_keys.append(eval_key)

        result_id = run_doc.get("results", None)
        if result_id is not None:
            result_ids.append(result_id)

    if result_ids:
        _logger.info("Deleting %d run result(s)", len(result_ids))
        if not dry_run:
            _delete_run_results(result_ids)

    _logger.info("Deleting evaluations %s from dataset '%s'", eval_keys, name)
    if not dry_run:
        dataset_dict["evaluations"] = {}
        conn.datasets.replace_one({"name": name}, dataset_dict)


def _get_logger(dry_run=False):
    if dry_run:
        return _DryRunLoggerAdapter(logger, {})

    return logger


class _DryRunLoggerAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        msg = "(dry run) " + msg
        return msg, kwargs


def _get_result_ids(dataset_dict):
    result_ids = []

    for run_doc in dataset_dict.get("annotation_runs", {}).values():
        result_id = run_doc.get("results", None)
        if result_id is not None:
            result_ids.append(result_id)

    for run_doc in dataset_dict.get("brain_methods", {}).values():
        result_id = run_doc.get("results", None)
        if result_id is not None:
            result_ids.append(result_id)

    for run_doc in dataset_dict.get("evaluations", {}).values():
        result_id = run_doc.get("results", None)
        if result_id is not None:
            result_ids.append(result_id)

    return result_ids


def _delete_run_results(result_ids):
    conn = get_db_conn()

    # Delete from GridFS
    conn.fs.files.delete_many({"_id": {"$in": result_ids}})
    conn.fs.chunks.delete_many({"files_id": {"$in": result_ids}})



"""
FiftyOne Server context

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import starlette.requests as strq
import starlette.responses as strp
import strawberry.asgi as gqla


import fiftyone as fo
from fiftyone.core.odm import get_async_db_client

from fiftyone.server.data import Context
from fiftyone.server.dataloader import dataloaders, get_dataloader


class GraphQL(gqla.GraphQL):
    async def get_context(
        self, request: strq.Request, response: strp.Response
    ) -> Context:
        db_client = get_async_db_client()
        db = db_client[fo.config.database_name]
        session = await db_client.start_session()
        loaders = {}
        for cls, config in dataloaders.items():
            loaders[cls] = get_dataloader(cls, config, db, session)

        return Context(
            db=db,
            session=session,
            dataloaders=loaders,
            request=request,
            response=response,
        )



"""
Annotation runs framework.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from fiftyone.core.runs import Run, RunConfig, RunInfo, RunResults


class AnnotationInfo(RunInfo):
    """Information about an annotation run on a dataset.

    Args:
        key: the annotation key
        timestamp (None): the UTC ``datetime`` when the annotation run was
            initiated
        config (None): the :class:`AnnotationMethodConfig` for the run
    """

    @classmethod
    def config_cls(cls):
        return AnnotationMethodConfig


class AnnotationMethodConfig(RunConfig):
    """Base class for configuring :class:`AnnotationMethod` instances.

    Args:
        **kwargs: any leftover keyword arguments after subclasses have done
            their parsing
    """

    pass


class AnnotationMethod(Run):
    """Base class for annotation methods.

    Args:
        config: an :class:`AnnotationMethodConfig`
    """

    @classmethod
    def run_info_cls(cls):
        return AnnotationInfo

    @classmethod
    def _runs_field(cls):
        return "annotation_runs"

    @classmethod
    def _run_str(cls):
        return "annotation run"

    @classmethod
    def _results_cache_field(cls):
        return "_annotation_cache"


class AnnotationResults(RunResults):
    """Base class for annotation run results."""

    pass



"""
Simple utility to download images from Flickr based on a text query.

Requires a user-specified API key, which can be obtained for free at
https://www.flickr.com/services/apps/create.

Copyright 2017-2022, Voxel51, Inc.
voxel51.com
"""
import argparse
from itertools import takewhile
import os

import flickrapi

import eta.core.storage as etas


def query_flickr(
    key, secret, query, number=50, path="data", query_in_path=True
):
    # Flickr api access key
    flickr = flickrapi.FlickrAPI(key, secret, cache=True)

    # could also query by tags and tag_mode='all'
    photos = flickr.walk(
        text=query, extras="url_c", per_page=50, sort="relevance"
    )

    urls = []
    for photo in takewhile(lambda _: len(urls) < number, photos):
        url = photo.get("url_c")
        if url is not None:
            urls.append(url)

    if query_in_path:
        basedir = os.path.join(path, query)
    else:
        basedir = path

    print(
        "Downloading %d images matching query '%s' to '%s'"
        % (len(urls), query, basedir)
    )
    client = etas.HTTPStorageClient()
    for url in urls:
        outpath = os.path.join(basedir, client.get_filename(url))
        client.download(url, outpath)
        print("Downloading image to '%s'" % outpath)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(add_help=True)

    parser.add_argument("key", type=str, help="Flickr API key")
    parser.add_argument("secret", type=str, help="Secret to Flickr API key")
    parser.add_argument("query", type=str, help="Query string to use")
    parser.add_argument(
        "-n",
        "--number",
        type=int,
        default=50,
        help="number of images to download (default: 50)",
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        default="data",
        help="path to download the images (created if needed)",
    )
    parser.add_argument(
        "--query-in-path", "-i", dest="query_in_path", action="store_true"
    )
    parser.add_argument(
        "--no-query-in-path", dest="query_in_path", action="store_false"
    )
    parser.set_defaults(query_in_path=True)

    args = parser.parse_args()

    query_flickr(
        key=args.key,
        secret=args.secret,
        query=args.query,
        number=args.number,
        path=args.path,
        query_in_path=args.query_in_path,
    )
